"""
 -----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Gravity model calibration
 Purpose:    Implement a procedure to calibrate gravity models

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    22/10/2016
 Updated:    02/10/2017
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------
"""
# The procedures implemented in this code are some of those suggested in
# Modelling Transport, 4th Edition
# Ortuzar and Willumsen, Wiley 2011
# The referred authors have no responsability over this work, of course

from time import clock

import numpy as np

from .gravity_application import GravityApplication, SyntheticGravityModel
from ..matrix import AequilibraeMatrix, AequilibraEData
from ..parameters import Parameters


class GravityCalibration:
    """"
        where function is: 'EXPO' or 'POWER'. 'GAMMA' and 'FRICTION FACTORS' to be implemented at a later time
        parameters are: 'max trip length'
        """

    def __init__(self, **kwargs):

        self.__required_parameters = ["max trip length", "max iterations", "max error"]
        self.parameters = kwargs.get("parameters", self.get_parameters())

        self.nan_as_zero = kwargs.get("nan_as_zero", False)
        self.matrix = kwargs.get("matrix")
        self.impedance = kwargs.get("impedance")
        deterrence_function = str(kwargs.get("function", "")).upper()

        if self.nan_as_zero:
            self.matrix = self.matrix.copy()
            self.impedance = self.impedance.copy()

        self.result_matrix = None
        self.rows = None
        self.columns = None
        self.gap = np.inf

        self.error = None
        self.gravity = None

        self.comput_core = None
        self.impedance_core = None

        self.itera = 0
        self.max_iter = None
        self.max_error = None
        self.gap = np.inf

        self.report = ["  #####    GRAVITY CALIBRATION    #####  ", ""]
        self.report.append("Functional form: " + deterrence_function)
        self.model = SyntheticGravityModel()
        if deterrence_function not in self.model.valid_functions:
            raise ValueError("Function needs to be one of these: " + ", ".join(self.model.valid_functions))
        else:
            self.model.function = deterrence_function

    def assemble_model(self, b1):
        # NEED TO SET PARAMETERS #
        if self.model.function == "EXPO":
            self.model.beta = float(b1)
        elif self.model.function == "POWER":
            self.model.alpha = float(b1)

    def calibrate(self):
        t = clock()
        # initialize auxiliary variables
        max_cost = self.parameters["max trip length"]
        self.max_iter = self.parameters["max iterations"]
        self.max_error = self.parameters["max error"]

        # Check the inputs
        self.check_inputs()
        if self.model.function in ["EXPO", "POWER"]:
            # filtering for all costs over limit

            a = 1
            if max_cost > 0:
                a = (self.impedance.matrix[self.impedance_core][:, :] < max_cost).astype(int)

            # weighted average cost
            self.report.append("Iteration: 1")
            cstar = np.nansum(
                self.impedance.matrix[self.impedance_core][:, :] * self.result_matrix.gravity[:, :] * a
            ) / np.nansum(self.result_matrix.gravity[:, :] * a)

            b0 = 1 / cstar

            self.assemble_model(b0)
            c0 = self.apply_gravity()
            for i in self.gravity.report:
                self.report.append("       " + i)
            self.report.append("")
            self.report.append("")

            bm1 = b0
            bm = b0 * c0 / cstar

            self.report.append("Iteration: 2")
            self.assemble_model(bm)

            cm = self.apply_gravity()
            for i in self.gravity.report:
                self.report.append("       " + i)
            self.report.append("Error: " + "{:.2E}".format(float(np.nansum(abs((bm / bm1) - 1)))))
            self.report.append("")
            cm1 = c0

        # While the max iterations has not been reached and the error is still too large
        self.itera = 2
        while self.itera < self.max_iter and self.gap > self.max_error:
            self.report.append("Iteration: " + str(self.itera + 1))
            aux = bm
            bm = ((cstar - cm1) * bm - (cstar - cm) * bm) / (cm - cm1)
            bm1 = aux
            cm1 = cm

            self.assemble_model(bm1)
            cm = self.apply_gravity()

            for i in self.gravity.report:
                self.report.append("       " + i)
            self.report.append("Error: " + "{:.2E}".format(float(np.nansum(abs((bm / bm1) - 1)))))
            self.report.append("")

            # compute convergence criteria
            self.gap = abs((bm / bm1) - 1)
            self.itera += 1

        if self.itera == self.max_iter:
            self.report.append(
                "DID NOT CONVERGE. Stopped in  " + str(self.itera) + "  with a global error of " + str(self.gap)
            )
        else:
            self.report.append("Converged in " + str(self.itera) + "  iterations to a global error of " + str(self.gap))
        s = clock() - t
        m, s1 = divmod(s, 60)
        s -= m * 60
        h, m = divmod(m, 60)
        t = "%d:%02d:%2.4f" % (h, m, s)

        self.report.append("Running time: " + t)

    def check_inputs(self):
        if not isinstance(self.impedance, AequilibraeMatrix):
            raise TypeError("Impedance matrix needs to be an instance of AequilibraEMatrix")

        if not isinstance(self.matrix, AequilibraeMatrix):
            raise TypeError("Observed matrix needs to be an instance of AequilibraEMatrix")

        # Check data dimensions
        if not np.array_equal(self.impedance.index, self.impedance.index):
            raise ValueError("Indices from impedance matrix do not match those from seed matrix")

        # Check if matrices were set for computation
        mats = [(self.matrix, "Observed matrix"), (self.impedance, "Impedance matrix")]
        for matrix, title in mats:
            if matrix.matrix_view is None:
                raise ValueError(title + " needs to be set for computation")
            else:
                if len(matrix.matrix_view.shape[:]) > 2:
                    raise ValueError(title + "' computational view needs to be set for a single matrix core")

            if np.nansum(matrix.matrix_view.data) == 0:
                raise ValueError(title + "has only zero values")
            if np.nanmin(matrix.matrix_view.data) < 0:
                raise ValueError(title + "has negative values")

        # Augment parameters if we happen to have only passed one
        default_parameters = self.get_parameters()
        for para in self.__required_parameters:
            if para not in self.parameters:
                self.parameters[para] = default_parameters[para]

        # Prepare the data for computation
        self.comput_core = self.matrix.view_names[0]

        self.result_matrix = self.matrix.copy(cores=[self.comput_core], names=["gravity"])

        self.rows = AequilibraEData()
        self.rows.create_empty(entries=self.matrix.zones, field_names=["rows"], memory_mode=True)
        self.rows.index[:] = self.matrix.index[:]
        self.rows.rows[:] = self.matrix.rows()[:]

        self.columns = AequilibraEData()
        self.columns.create_empty(entries=self.matrix.zones, field_names=["columns"], memory_mode=True)
        self.columns.index[:] = self.matrix.index[:]
        self.columns.columns[:] = self.matrix.columns()[:]

        self.impedance_core = self.impedance.view_names[0]

    def apply_gravity(self):
        args = {
            "impedance": self.impedance,
            "rows": self.rows,
            "row_field": "rows",
            "columns": self.columns,
            "column_field": "columns",
            "model": self.model,
            "parameters": self.parameters,
            "nan_as_zero": self.nan_as_zero,
        }

        self.gravity = GravityApplication(**args)
        self.gravity.apply()
        self.result_matrix = self.gravity.output

        return np.nansum(
            self.impedance.matrix[self.impedance_core][:, :] * self.result_matrix.gravity[:, :]
        ) / np.nansum(self.result_matrix.gravity[:, :])

    def get_parameters(self):
        par = Parameters().parameters
        para = par["distribution"]["ipf"].copy()
        para.update(par["distribution"]["gravity"])
        return para



import os
import yaml


class Parameters:
    """
    Parameters are used in many procedures, and are often defined only in thi parameters.yml file ONLY
    Parameters are organized in the following groups:

    assignment:
    distribution:

    system:
    cpus: Maximum threads to be used in any procedure
    default_directory: If is the directory QGIS file opening/saving dialogs will try to open as standard
    driving side: For purposes of plotting on QGIS
    logging: Level of logging to be written to temp/aequilibrae.log: Levels are those from the Python logging library
                0: 'NOTSET'
                10: 'DEBUG'
                20: 'INFO'
                30: 'WARNING'
                40: 'ERROR'
                50: 'CRITICAL'
            both numeric and text accepted
    report zeros:
    temp directory:

    """

    def __init__(self):
        self.path = os.path.dirname(os.path.realpath(__file__))

        file = os.path.join(self.path, "parameters.yml")
        with open(file, "r") as yml:
            self.parameters = yaml.load(yml, Loader=yaml.SafeLoader)

    def write_back(self):
        stream = open(self.path + "/parameters.yaml", "w")
        yaml.dump(self.parameters, stream, default_flow_style=False)
        stream.close()



from unittest import TestCase

import numpy as np

from aequilibrae.distribution import Ipf
from aequilibrae.matrix import AequilibraEData
from aequilibrae.matrix import AequilibraeMatrix

zones = 100

# row vector
args = {"entries": zones, "field_names": ["rows"], "data_types": [np.float64], "memory_mode": True}
row_vector = AequilibraEData()
row_vector.create_empty(**args)
row_vector.rows[:] = np.random.rand(zones)[:] * 1000
row_vector.index[:] = np.arange(zones)[:]
# column vector
args["field_names"] = ["columns"]
column_vector = AequilibraEData()
column_vector.create_empty(**args)
column_vector.columns[:] = np.random.rand(zones)[:] * 1000
column_vector.index[:] = np.arange(zones)[:]
# balance vectors
column_vector.columns[:] = column_vector.columns[:] * (row_vector.rows.sum() / column_vector.columns.sum())

# seed matrix_procedures
name_test = AequilibraeMatrix().random_name()
args = {"file_name": name_test, "zones": zones, "matrix_names": ["seed"]}

matrix = AequilibraeMatrix()
matrix.create_empty(**args)
matrix.seed[:, :] = np.random.rand(zones, zones)[:, :]
matrix.computational_view(["seed"])
matrix.matrix_view[1, 1] = np.nan
matrix.index[:] = np.arange(zones)[:]


class TestIpf(TestCase):
    def test_fit(self):
        # The IPF per se
        args = {
            "matrix": matrix,
            "rows": row_vector,
            "row_field": "rows",
            "columns": column_vector,
            "column_field": "columns",
            "nan_as_zero": False,
        }

        fratar = Ipf(**args)
        fratar.fit()

        result = fratar.output
        if (np.nansum(result.matrix_view) - np.nansum(row_vector.data["rows"])) > 0.001:
            print(fratar.gap)
            for f in fratar.report:
                print(f)
            self.fail("Ipf did not converge")

        if fratar.gap > fratar.parameters["convergence level"]:
            print(fratar.gap)
            for f in fratar.report:
                print(f)
            self.fail("Ipf did not converge")



import os
import sqlite3
import tempfile
import uuid

import numpy as np
from numpy.lib.format import open_memmap

""""""
"""
 -----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       AequilibraE Database
 Purpose:    Implements a class to represent flat databases with an unique id field.
             It is really just a wrapper around numpy recarray memmap

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    2017-10-02
 Updated:
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------
 """

MEMORY = 1
DISK = 0
data_export_types = ["aed", "csv", "sqlite"]


class AequilibraEData(object):
    def __init__(self):
        self.data = None
        self.file_path = None
        self.entries = None
        self.fields = None
        self.num_fields = None
        self.data_types = None
        self.aeq_index_type = None
        self.memory_mode = None

    def create_empty(self, file_path=None, entries=1, field_names=None, data_types=None, memory_mode=False):
        """
        :param file_path: Optional. Full path for the output data file. If *memory_false* is 'false' and path is missing,
        then the file is created in the temp folder
        :param entries: Number of records in the dataset. Default is 1
        :param field_names: List of field names for this dataset. If no list is provided, the field 'data' will be created
        :param data_types: List of data types for the dataset. Types need to be NumPy data types (e.g. np.int16,
        np.float64). If no list of types are provided, type will be *np.float64*
        :param memory_mode: If true, dataset will be kept in memory. If false, the dataset will  be a memory-mapped numpy array
        :return: # nothing. Associates a dataset with the AequilibraEData object
        """

        if file_path is not None or memory_mode:
            if field_names is None:
                field_names = ["data"]

            if data_types is None:
                data_types = [np.float64] * len(field_names)

            self.file_path = file_path
            self.entries = entries
            self.fields = field_names
            self.data_types = data_types
            self.aeq_index_type = np.uint64

            if memory_mode:
                self.memory_mode = MEMORY
            else:
                self.memory_mode = DISK
                if self.file_path is None:
                    self.file_path = self.random_name()

            # Consistency checks
            if not isinstance(self.fields, list):
                raise ValueError('Titles for fields, "field_names", needs to be a list')

            if not isinstance(self.data_types, list):
                raise ValueError('Data types, "data_types", needs to be a list')
            # The check below is not working properly with the QGIS importer
            # else:
            #     for dt in self.data_types:
            #         if not isinstance(dt, type):
            #             raise ValueError('Data types need to be Python or Numpy data types')

            for field in self.fields:
                if field in object.__dict__:
                    raise Exception(field + " is a reserved name. You cannot use it as a field name")

            self.num_fields = len(self.fields)

            dtype = [("index", self.aeq_index_type)]
            dtype.extend([(self.fields[i], self.data_types[i]) for i in range(self.num_fields)])

            # the file
            if self.memory_mode:
                self.data = np.recarray((self.entries,), dtype=dtype)
            else:
                self.data = open_memmap(self.file_path, mode="w+", dtype=dtype, shape=(self.entries,))

    def __getattr__(self, field_name):

        if field_name in object.__dict__:
            return self.__dict__[field_name]

        if self.data is not None:
            if field_name in self.fields:
                return self.data[field_name]

            if field_name == "index":
                return self.data["index"]

            raise AttributeError("No such method or data field! --> " + str(field_name))
        else:
            raise AttributeError("Data container is empty")

    def load(self, file_path):
        """
        :param file_path: Full file path to the AequilibraEDataset to be loaded
        :return: Loads the dataset into the AequilibraEData instance
        """
        f = open(file_path)
        self.file_path = os.path.realpath(f.name)
        f.close()

        # Map in memory and load data names plus dimensions
        self.data = open_memmap(self.file_path, mode="r+")

        self.entries = self.data.shape[0]
        self.fields = [x for x in self.data.dtype.fields if x != "index"]
        self.num_fields = len(self.fields)
        self.data_types = [self.data[x].dtype.type for x in self.fields]

    def export(self, file_name, table_name="aequilibrae_table"):
        """
        :param file_name: File name with PATH and extension (csv, or sqlite3, sqlite or db)
        :param table_name: It only applies if you are saving to an SQLite table. Otherwise ignored
        :return:
        """

        file_type = os.path.splitext(file_name)[1]
        headers = ["index"]
        headers.extend(self.fields)

        if file_type == ".csv":
            fmt = "%d"
            for dt in self.data_types:
                if np.issubdtype(dt, np.floating):
                    fmt += ",%f"
                elif np.issubdtype(dt, np.integer):
                    fmt += ",%d"
            np.savetxt(
                file_name, self.data[np.newaxis, :][0], delimiter=",", fmt=fmt, header=",".join(headers), comments=""
            )

        elif file_type in [".sqlite", ".sqlite3", ".db"]:
            # Connecting to the database file
            conn = sqlite3.connect(file_name)
            c = conn.cursor()
            # Creating the table, but before deletes if the table exists
            c.execute("""DROP TABLE IF EXISTS """ + table_name)
            fi = ""
            qm = "?"
            for f in headers[1:]:
                fi += ", " + f + " REAL"
                qm += ", ?"

            c.execute("""CREATE TABLE """ + table_name + """ (link_id INTEGER PRIMARY KEY""" + fi + ")" "")
            c.execute("BEGIN TRANSACTION")
            c.executemany("INSERT INTO " + table_name + " VALUES (" + qm + ")", self.data)
            c.execute("END TRANSACTION")
            conn.commit()
            conn.close()

    @staticmethod
    def random_name():
        return os.path.join(tempfile.gettempdir(), "Aequilibrae_data_" + str(uuid.uuid4()) + ".aed")



class Route:
    """
    Represents each one of the routes in a GTFS dataset (from https://developers.google.com/transit/gtfs/reference/)

    Fields
    ______

    * **id** `(route_id)` **Required** The route_id field contains an ID that uniquely identifies a route. The route_id is dataset unique.
    * **agency_id** `(agency_id)` **Optional** The agency_id field defines an agency for the specified route. This value is referenced from the agency.txt file. Use this field when you are providing data for routes from more than one agency.
    * **short_name** `(route_short_name)` **Required** The route_short_name contains the short name of a route. This will often be a short, abstract identifier like "32", "100X", or "Green" that riders use to identify a route, but which doesn't give any indication of what places the route serves. At least one of route_short_name or route_long_name must be specified, or potentially both if appropriate. If the route does not have a short name, please specify a route_long_name and use an empty string as the value for this field.
    * **long_name** `(route_long_name)` **Required** The route_long_name contains the full name of a route. This name is generally more descriptive than the route_short_name and will often include the route's destination or stop. At least one of route_short_name or route_long_name must be specified, or potentially both if appropriate. If the route does not have a long name, please specify a route_short_name and use an empty string as the value for this field.
    * **desc** `(route_desc)` **Optional** The route_desc field contains a description of a route. Please provide useful, quality information. Do not simply duplicate the name of the route. For example, "A trains operate between Inwood-207 St, Manhattan and Far Rockaway-Mott Avenue, Queens at all times. Also from about 6AM until about midnight, additional A trains operate between Inwood-207 St and Lefferts Boulevard (trains typically alternate between Lefferts Blvd and Far Rockaway)."
    * **type** `(route_type)` **Required** The route_type field describes the type of transportation used on a route. Valid values for this field are:

            * 0 - Tram, Streetcar, Light rail. Any light rail or street level system within a metropolitan area.
            * 1 - Subway, Metro. Any underground rail system within a metropolitan area.
            * 2 - Rail. Used for intercity or long-distance travel.
            * 3 - Bus. Used for short- and long-distance bus routes.
            * 4 - Ferry. Used for short- and long-distance boat service.
            * 5 - Cable car. Used for street-level cable cars where the cable runs beneath the car.
            * 6 - Gondola, Suspended cable car. Typically used for aerial cable cars where the car is suspended from the cable.
            * 7 - Funicular. Any rail system designed for steep inclines.
    * **url** `(route_url)` **Optional** The route_url field contains the URL of a web page about that particular route. This should be different from the agency_url. The value must be a fully qualified URL that includes http:// or https://, and any special characters in the URL must be correctly escaped. See http://www.w3.org/Addressing/URL/4_URI_Recommentations.html for a description of how to create fully qualified URL values.
    * **color** `(route_color)` **Optional** In systems that have colors assigned to routes, the route_color field defines a color that corresponds to a route. The color must be provided as a six-character hexadecimal number, for example, 00FFFF. If no color is specified, the default route color is white (FFFFFF). The color difference between route_color and route_text_color should provide sufficient contrast when viewed on a black and white screen. The W3C Techniques for Accessibility Evaluation And Repair Tools document offers a useful algorithm for evaluating color contrast. There are also helpful online tools for choosing contrasting colors, including the snook.ca Color Contrast Check application.
    * **text_color** `(route_text_color)` **Optional** The route_text_color field can be used to specify a legible color to use for text drawn against a background of route_color. The color must be provided as a six-character hexadecimal number, for example, FFD700. If no color is specified, the default text color is black (000000). The color difference between route_color and route_text_color should provide sufficient contrast when viewed on a black and white screen.
    * **sort_order** `(route_sort_order)` **Optional** The route_sort_order field can be used to order the routes in a way which is ideal for presentation to customers. It must be a non-negative integer. Routes with smaller route_sort_order values should be displayed before routes with larger route_sort_order values.

    # Routes are associated with trips
    """

    def __init__(self):
        # Fields from the routes.txt
        self.id = None
        self.agency_id = None
        self.short_name = None
        self.long_name = None
        self.desc = None
        self.type = None
        self.url = None
        self.color = None
        self.text_color = None
        self.sort_order = None

        # Will contain the list of all trips for this route
        self.trips = []

        # Will contain all shapes for the route
        self.shapes = None

        # Will contain the list of all the stops in the stop
        self.stops = []



# coding=utf-8
class Agency:
    """
    contents of the **agency.txt** file  (from https://developers.google.com/transit/gtfs/reference/)

    Fields
    ______

    * **id** `(agency_id)` **Optional** - The agency_id field is an ID that uniquely identifies a transit agency. A transit feed may represent data from more than one agency. The agency_id is dataset unique. This field is optional for transit feeds that only contain data for a single agency.

    * **name** `(agency_name)` **Required** - The agency_name field contains the full name of the transit agency. Google Maps will display this name.

    * **url** `(agency_url)` **Required** - The agency_url field contains the URL of the transit agency. The value must be a fully qualified URL that includes http:// or https://, and any special characters in the URL must be correctly escaped. See http://www.w3.org/Addressing/URL/4_URI_Recommentations.html for a description of how to create fully qualified URL values.

    * **timezone** `(agency_timezone)` **Required** - The agency_timezone field contains the timezone where the transit agency is located. Timezone names never contain the space character but may contain an underscore. Please refer to http://en.wikipedia.org/wiki/List_of_tz_zones for a list of valid values. If multiple agencies are specified in the feed, each must have the same agency_timezone.

    * **lang** `(agency_lang)` **Optional** - The agency_lang field contains a two-letter ISO 639-1 code for the primary language used by this transit agency. The language code is case-insensitive (both en and EN are accepted). This setting defines capitalization rules and other language-specific settings for all text contained in this transit agency's feed. Please refer to http://www.loc.gov/standards/iso639-2/php/code_list.php for a list of valid values.

    * **phone** `(agency_phone)` **Optional** - The agency_phone field contains a single voice telephone number for the specified agency. This field is a string value that presents the telephone number as typical for the agency's service area. It can and should contain punctuation marks to group the digits of the number. Dialable text (for example, TriMet's "503-238-RIDE") is permitted, but the field must not contain any other descriptive text.

    * **fare_url** `(agency_url)` **Optional** - The agency_fare_url specifies the URL of a web page that allows a rider to purchase tickets or other fare instruments for that agency online. The value must be a fully qualified URL that includes http:// or https://, and any special characters in the URL must be correctly escaped. See http://www.w3.org/Addressing/URL/4_URI_Recommentations.html for a description of how to create fully qualified URL values.

    * **email** `(agency_email)` **Optional** - Contains a single valid email address actively monitored by the agency’s customer service department. This email address will be considered a direct contact point where transit riders can reach a customer service representative at the agency.
    """

    def __init__(self):
        """
        Initializes the class with members corresponding to all fields in the GTFS specification. See Stop class
        documentation
        """
        self.id = None
        self.name = None
        self.url = None
        self.timezone = None
        self.lang = None
        self.phone = None
        self.fare_url = None
        self.email = None



class CalendarDates:
    """
    The calendar_dates table allows you to explicitly activate or disable service IDs by date. You can use it in two ways.

    File: **Optional**

    Use
    ______

    Recommended: Use calendar_dates.txt in conjunction with calendar.txt, where calendar_dates.txt defines any exceptions to the default service categories defined in the calendar.txt file. If your service is generally regular, with a few changes on explicit dates (for example, to accommodate special event services, or a school schedule), this is a good approach.
    Alternate: Omit calendar.txt, and include ALL dates of service in calendar_dates.txt. If your schedule varies most days of the month, or you want to programmatically output service dates without specifying a normal weekly schedule, this approach may be preferable.

    Fields
    ______

    * **service_id** `(service_id)`  **Required** - The service_id contains an ID that uniquely identifies a set of dates when a service exception is available for one or more routes. Each (service_id, date) pair can only appear once in calendar_dates.txt. If the a service_id value appears in both the calendar.txt and calendar_dates.txt files, the information in calendar_dates.txt modifies the service information specified in calendar.txt. This field is referenced by the trips.txt file.
    * **date** `(date)`  **Required** - The date field specifies a particular date when service availability is different than the norm. You can use the exception_type field to indicate whether service is available on the specified date. The date field's value should be in YYYYMMDD format.
    * **exception_type** `(exception_type)`  **Required** - The exception_type indicates whether service is available on the date specified in the date field.

            * A value of 1 indicates that service has been added for the specified date.
            * A value of 2 indicates that service has been removed for the specified date.

    Example
    ______

    Suppose a route has one set of trips available on holidays and another set of trips available on all other days. You could have one service_id that corresponds to the regular service schedule and another service_id that corresponds to the holiday schedule. For a particular holiday, you would use the calendar_dates.txt file to add the holiday to the holiday service_id and to remove the holiday from the regular service_id schedule.

    """

    def __init__(self):
        self.service_id = []
        self.date = []
        self.exception_type = []



"""
-----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Traffic assignment
 Purpose:    Implement traffic assignment algorithms based on Cython's network loading procedures

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    15/09/2013
 Updated:    2018-07-01
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
-----------------------------------------------------------------------------------------------------------
 """

import importlib
import sys
import threading
from multiprocessing.dummy import Pool as ThreadPool

import numpy as np

from .AoN import one_to_all
from .multi_threaded_aon import MultiThreadedAoN
from ..utils import WorkerThread

have_pyqt5 = importlib.util.find_spec("PyQt5")
if have_pyqt5 is None:
    pyqt = False
else:
    from PyQt5.QtCore import pyqtSignal as SIGNAL

    pyqt = True

sys.dont_write_bytecode = True


class allOrNothing(WorkerThread):
    assignment = SIGNAL(object)

    def __init__(self, matrix, graph, results):
        WorkerThread.__init__(self, None)

        self.matrix = matrix
        self.graph = graph
        self.results = results
        self.aux_res = MultiThreadedAoN()
        self.report = []
        self.cumulative = 0

        if results.__graph_id__ != graph.__id__:
            raise ValueError("Results object not prepared. Use --> results.prepare(graph)")

        if results.__graph_id__ is None:
            raise ValueError("The results object was not prepared. Use results.prepare(graph)")

        elif results.__graph_id__ != graph.__id__:
            raise ValueError("The results object was prepared for a different graph")

        elif matrix.matrix_view is None:
            raise ValueError(
                "Matrix was not prepared for assignment. "
                "Please create a matrix_procedures view with all classes you want to assign"
            )

        elif not np.array_equal(matrix.index, graph.centroids):
            raise ValueError("Matrix and graph do not have compatible set of centroids.")

    def doWork(self):
        self.execute()

    def execute(self):
        if pyqt:
            self.assignment.emit(["zones finalized", 0])

        self.aux_res.prepare(self.graph, self.results)
        self.matrix.matrix_view = self.matrix.matrix_view.reshape(
            (self.graph.num_zones, self.graph.num_zones, self.results.classes["number"])
        )
        mat = self.matrix.matrix_view
        pool = ThreadPool(self.results.cores)
        all_threads = {"count": 0}
        for orig in self.matrix.index:
            i = int(self.graph.nodes_to_indices[orig])
            if np.nansum(mat[i, :, :]) > 0:
                if self.graph.fs[i] == self.graph.fs[i + 1]:
                    self.report.append("Centroid " + str(orig) + " is not connected")
                else:
                    pool.apply_async(self.func_assig_thread, args=(orig, all_threads))
                    # one_to_all(orig, self.matrix, self.graph, self.results, self.aux_res, 0)
        pool.close()
        pool.join()
        self.results.link_loads = np.sum(self.aux_res.temp_link_loads, axis=2)

        if pyqt:
            self.assignment.emit(["text AoN", "Saving Outputs"])
            self.assignment.emit(["finished_threaded_procedure", None])

    def func_assig_thread(self, O, all_threads):
        if threading.get_ident() in all_threads:
            th = all_threads[threading.get_ident()]
        else:
            all_threads[threading.get_ident()] = all_threads["count"]
            th = all_threads["count"]
            all_threads["count"] += 1
        x = one_to_all(O, self.matrix, self.graph, self.results, self.aux_res, th)
        self.cumulative += 1
        if x != O:
            self.report.append(x)
        if pyqt:
            self.assignment.emit(["zones finalized", self.cumulative])
            txt = str(self.cumulative) + " / " + str(self.matrix.zones)
            self.assignment.emit(["text AoN", txt])



import os

with open(os.path.join(os.path.dirname(__file__), "parameters.pxi"), "r") as a:
    for i in a.readlines():
        if "VERSION" in i:
            version = i[10:].rstrip()
        if "MINOR_VRSN" in i:
            minor_version = i[13:].rstrip()
        if "release_name" in i:
            release_name = i[16:-1].rstrip()
        if "binary" in i:
            binary_version = i[17:-1].rstrip()

release_version = "{}.{}".format(version.rstrip(), minor_version.rstrip())



import os
import sys
import tempfile
import logging
from .parameters import Parameters

sys.dont_write_bytecode = True


def StartsLogging():
    # CREATE THE LOGGER
    temp_folder = Parameters().parameters["system"]["temp directory"]
    if not os.path.isdir(temp_folder):
        temp_folder = tempfile.gettempdir()

    log_file = os.path.join(temp_folder, "aequilibrae.log")
    if not os.path.isfile(log_file):
        a = open(log_file, "w")
        a.close()

    logger = logging.getLogger("aequilibrae")
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    if not len(logger.handlers):
        ch = logging.FileHandler(log_file)
        ch.setFormatter(formatter)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)



from unittest import TestCase
import os
import tempfile
import numpy as np
from aequilibrae.paths import Graph
from os.path import join
import sys
from .parameters_test import centroids

# Adds the folder with the data to the path and collects the paths to the files
# lib_path = os.path.abspath(os.path.join('..', '../tests'))
# sys.path.append(lib_path)
from ...data import path_test, test_graph, test_network


class TestGraph(TestCase):
    def test_create_from_geography(self):
        self.graph = Graph()
        self.graph.create_from_geography(
            test_network,
            "link_id",
            "dir",
            "distance",
            centroids=centroids,
            skim_fields=[],
            anode="A_NODE",
            bnode="B_NODE",
        )
        self.graph.set_graph(cost_field="distance", block_centroid_flows=True)

    def test_load_network_from_csv(self):
        pass

    def test_prepare_graph(self):
        self.test_create_from_geography()
        self.graph.prepare_graph(centroids)

        reference_graph = Graph()
        reference_graph.load_from_disk(test_graph)
        if not np.array_equal(self.graph.graph, reference_graph.graph):
            self.fail("Reference graph and newly-prepared graph are not equal")

    def test_set_graph(self):
        self.test_prepare_graph()
        self.graph.set_graph(cost_field="distance", block_centroid_flows=True)
        if self.graph.num_zones != centroids.shape[0]:
            self.fail("Number of centroids not properly set")
        if self.graph.num_links != 222:
            self.fail("Number of links not properly set")
        if self.graph.num_nodes != 93:
            self.fail("Number of nodes not properly set - " + str(self.graph.num_nodes))

    def test_save_to_disk(self):
        self.test_create_from_geography()
        self.graph.save_to_disk(join(path_test, "aequilibrae_test_graph.aeg"))
        self.graph_id = self.graph.__id__
        self.graph_version = self.graph.__version__

    def test_load_from_disk(self):
        self.test_save_to_disk()
        reference_graph = Graph()
        reference_graph.load_from_disk(test_graph)

        new_graph = Graph()
        new_graph.load_from_disk(join(path_test, "aequilibrae_test_graph.aeg"))

        comparisons = [
            ("Graph", new_graph.graph, reference_graph.graph),
            ("b_nodes", new_graph.b_node, reference_graph.b_node),
            ("Forward-Star", new_graph.fs, reference_graph.fs),
            ("cost", new_graph.cost, reference_graph.cost),
            ("centroids", new_graph.centroids, reference_graph.centroids),
            ("skims", new_graph.skims, reference_graph.skims),
            ("link ids", new_graph.ids, reference_graph.ids),
            ("Network", new_graph.network, reference_graph.network),
            ("All Nodes", new_graph.all_nodes, reference_graph.all_nodes),
            (
                "Nodes to indices",
                new_graph.nodes_to_indices,
                reference_graph.nodes_to_indices,
            ),
        ]

        for comparison, newg, refg in comparisons:
            if not np.array_equal(newg, refg):
                self.fail(
                    "Reference %s and %s created and saved to disk are not equal"
                    % (comparison, comparison)
                )

        comparisons = [
            ("nodes", new_graph.num_nodes, reference_graph.num_nodes),
            ("links", new_graph.num_links, reference_graph.num_links),
            ("zones", new_graph.num_zones, reference_graph.num_zones),
            (
                "block through centroids",
                new_graph.block_centroid_flows,
                reference_graph.block_centroid_flows,
            ),
            ("Graph ID", new_graph.__id__, self.graph_id),
            ("Graph Version", new_graph.__version__, self.graph_version),
        ]

        for comparison, newg, refg in comparisons:
            if newg != refg:
                self.fail(
                    "Reference %s and %s created and saved to disk are not equal"
                    % (comparison, comparison)
                )

    def test_reset_single_fields(self):
        pass

    def test_add_single_field(self):
        pass

    def test_available_skims(self):
        self.test_set_graph()
        if self.graph.available_skims() != ["distance"]:
            self.fail("Skim availability with problems")



import os
import tempfile
from unittest import TestCase

import numpy as np

from aequilibrae.matrix import AequilibraEData

file_path = AequilibraEData().random_name()
args = {
    "file_path": file_path,
    "entries": 100,
    "field_names": ["d", "data2", "data3"],
    "data_types": [np.float64, np.float32, np.int8],
}


class TestAequilibraEData(TestCase):
    def test___init__(self):
        # Generates the dataset
        dataset = AequilibraEData()
        dataset.create_empty(**args)

        dataset.index[:] = np.arange(dataset.entries) + 100
        dataset.d[:] = dataset.index[:] ** 2
        if dataset.index[70] != 170:
            self.fail()

        if int(dataset.d[70]) != 28900:
            self.fail()

        # removes the dataset
        del dataset

    def test_load(self):
        # re-imports the dataset
        self.ad = AequilibraEData()
        self.ad.load(file_path)

        # checks if the values were properly saved
        if self.ad.index[70] != 170:
            self.fail("Value for data index test was not as expected")

        if int(self.ad.d[70]) != 28900:
            self.fail("Value for data field test was not as expected")

        for f in self.ad.fields:
            if f not in args["field_names"]:
                self.fail("Could not retrieve all fields")

    def test_export(self):
        self.test_load()
        temp_name = os.path.join(tempfile.gettempdir(), "aequilibrae_data_example.csv")
        self.ad.export(temp_name)



import numpy as np
import csv
import copy


def parse_csv(file_name: str, column_order=[]):
    tot = []
    if isinstance(file_name, str):
        csvfile = open(file_name, encoding="utf-8-sig")
    else:
        csvfile = file_name

    contents = csv.reader(csvfile, delimiter=",", quotechar='"')
    for row in contents:
        tot.append(row)
    titles = tot.pop(0)
    if tot:
        data = np.core.records.fromrecords(tot, names=titles)
    else:
        return None

    if column_order:
        col_names = [x for x in column_order.keys() if x in data.dtype.names]
        data = data[col_names]

        # Define sizes for the string variables
        column_order = copy.deepcopy(column_order)
        for c in col_names:
            if column_order[c] is str:
                if data[c].dtype.char.upper() == "S":
                    column_order[c] = data[c].dtype
                else:
                    column_order[c] = "S128"
            else:
                if data[c].dtype.char.upper() in ["U", "S"]:
                    data[c][data[c] == ""] = "0"

        new_data_dt = [(f, column_order[f]) for f in col_names]

        if int(data.shape.__len__()) > 0:
            return np.array(data, new_data_dt)
        else:
            return data
    else:
        return data



import importlib.util as iutil
import os
import tempfile
import uuid
from shutil import copyfile
from typing import List
import functools

import numpy as np

""""""
"""-----------------------------------------------------------------------------------------------------------
Package:    AequilibraE
Name:       AequilibraE Matrix
Purpose:    Implements a new class to represent multi-layer matrices
Original Author:  Pedro Camargo (c@margo.co)
Contributors: Pedro Camargo
Last edited by: Pedro Camargo
Website:    www.AequilibraE.com
Repository:  https://github.com/AequilibraE/AequilibraE
Created:    2017-10-02
Updated:    2018-07-08
Copyright:   (c) AequilibraE authors
Licence:     See LICENSE.TXT
-----------------------------------------------------------------------------------------------------------"""

# CONSTANTS
VERSION = 1  # VERSION OF THE MATRIX FORMAT
INT = 0
FLOAT = 1
COMPLEX = 2
CORE_NAME_MAX_LENGTH = 50
INDEX_NAME_MAX_LENGTH = 20
MATRIX_NAME_MAX_LENGTH = 20
MATRIX_DESCRIPTION_MAX_LENGTH = 144  # Why not a tweet?
NOT_COMPRESSED = 0
COMPRESSED = 1

# TODO:  Add an aggregate method
# TODO: Add check that the index 1 (or zero) has only unique values?


# Matrix structure
#
# What:  Version | Compress flag | # cells: compressed matrix | # of zones | # of cores | # of indices (Y) | Data type |
# Size:   uint8  |    uint8      |        uint64              |  uint32    |   uint8    |     uint8        |  uint8    |
# Shape:     1   |       1       |           1                |     1      |            |       1          |     1     |
# Offset:    0   |       1       |           2                |     10     |     14     |      15          |    16     |
#
#
# What:    Data size |  matrix name | matrix description | Core names |   index names    |
# Size:     uint8    |     S20      |          S144      |    S50     |      S20         |
# Shape:      1      |      1       |            1       |  [cores]   |    [indices]     |
# Offset:     17     |     18       |          38        |     182    |  182 + 50*cores  |
#
# What:         indices          |             Matrices                 |
# Size:         uint64           |      f(Data type, Data size)         |
# Shape:     [zones, indices]    |       [zones, zones, cores]          |
# Offset:  18 + 50*cores + Y*20  |   18 + 50*cores + Y*20 + Y*zones*8   |


matrix_export_types = ["Aequilibrae matrix (*.aem)", "Comma-separated file (*.csv)"]


class AequilibraeMatrix(object):
    """
    Matrix class
    """

    def __init__(self):
        """
        Creates a memory instance for a matrix, that can be used to load an existing matrix or to create an empty one
        """
        self.file_path = None
        self.dtype = None
        self.num_indices = None
        self.index_names = None
        self.compressed = NOT_COMPRESSED
        self.matrix_view = None
        self.view_names = None
        self.matrix_hash = {}
        self.index = None
        self.indices = None
        self.matrix = None
        self.matrices = None
        self.cores = None
        self.zones = None
        self.dtype = None
        self.names = None
        self.name = None
        self.description = None
        self.current_index = None
        self.__version__ = VERSION  # Writes file version

    def create_empty(
        self,
        file_name: str = None,
        zones: int = None,
        matrix_names: List[str] = None,
        data_type: np.dtype = np.float64,
        index_names: List[str] = None,
        compressed: bool = False,
    ):
        """
        Creates an empty matrix in the AequilibraE format

        Parameters
        ----------
        file_name: string
            Local path to the matrix file

        zones: integer
            Number of zones in the model (Integer). Maximum number of zones in a matrix is 4,294,967,296

        matrix_names: list
            A regular Python list of names of the matrix. Limit is 50 characters each. Maximum number of cores per
            matrix is 256

        data_type: np.dtype, optional
            Data type of the matrix as NUMPY data types (NP.int32, np.int64, np.float32, np.float64).
            Dafaultis np.float64

        index_names: list, optional
            A regular Python list of names for indices. Limit is 20 characters each).
            Maximum number of indices per matrix is 256

        compressed: bool, optional
            Whether it is a flat matrix or a compressed one(Boolean - Not yet implemented)

        ------------------------------------------------------------
        Example

        >>> zones_in_the_model = 3317
        >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

        >>> mat = AequilibraeMatrix()
        >>> mat.create_empty(file_name='my/path/to/file', zones=zones_in_the_model, matrix_names= names_list)
        >>> mat.num_indices
        1
        >>> mat.zones
        3317
        >>> np.sum(mat[trips])
        0.0
        """

        self.file_path = file_name
        self.zones = zones
        self.index_names = index_names
        self.dtype = data_type

        # Matrix compression still not supported
        if compressed:
            compressed = False
            raise Warning("Matrix compression not yet supported")

        if compressed:
            self.compressed = COMPRESSED
        else:
            self.compressed = NOT_COMPRESSED

        if index_names is None:
            self.index_names = ["main_index"]
        else:
            if isinstance(index_names, list) or isinstance(index_names, tuple):
                self.index_names = index_names
                for ind_name in index_names:
                    if isinstance(ind_name, str):
                        if len(ind_name) > INDEX_NAME_MAX_LENGTH:
                            raise ValueError(
                                "Index names need to be be shorter "
                                "than {}: {}".format(INDEX_NAME_MAX_LENGTH, ind_name)
                            )
                    else:
                        raise ValueError("Index names need to be strings: " + str(ind_name))
            else:
                raise Exception("Index names need to be provided as a list")
        self.num_indices = len(self.index_names)

        if matrix_names is None:
            matrix_names = ["mat"]
        else:
            if isinstance(matrix_names, list) or isinstance(matrix_names, tuple):
                for mat_name in matrix_names:
                    if isinstance(mat_name, str):
                        if mat_name in object.__dict__:
                            raise ValueError(mat_name + " is a reserved name")
                        if len(mat_name) > CORE_NAME_MAX_LENGTH:
                            raise ValueError(
                                "Matrix names need to be be shorter "
                                "than {}: {}".format(CORE_NAME_MAX_LENGTH, mat_name)
                            )
                    else:
                        raise ValueError("Matrix core names need to be strings: " + str(mat_name))
            else:
                raise Exception("Matrix names need to be provided as a list")

        self.names = [x for x in matrix_names]
        self.cores = len(self.names)
        if None not in [self.file_path, self.zones]:
            self.__write__()

    def create_from_omx(
        self,
        file_path: str,
        omx_path: str,
        cores: List[str] = None,
        mappings: List[str] = None,
        robust: bool = True,
        compressed: bool = False,
    ):
        """
        :param file_path: Path for the output AequilibraEMatrix
        :param omx_path: Path to the OMX file one wants to import
        :param cores: List of matrix cores to be imported
        :param mappings: List of the matrix mappings (i.e. indices, centroid numbers) to be imoprted
        :param: robust: Boolean for whether AequilibraE should try to adjust the names for cores and indices in case they are too long
        :param: compressed
        :return:
        """

        def robust_name(input_name: str, max_length: int, forbiden_names: List[str]) -> str:
            return_name = input_name
            if len(input_name) > max_length:
                return_name = input_name[:max_length]

            if return_name not in forbiden_names:
                return return_name

            for i in range(999999):
                x = len(str(i))

                if x + len(return_name) > max_length:
                    trial_name = return_name[: max_length - x] + str(i)
                else:
                    trial_name = return_name + str(i)

                if trial_name not in forbiden_names:
                    return trial_name

        spec = iutil.find_spec("openmatrix")
        if spec is None:
            print("Open Matrix is not installed. Cannot continue")
            return

        import openmatrix as omx

        src = omx.open_file(omx_path, "r")

        avail_cores = src.list_matrices()

        if cores is None:
            do_cores = avail_cores
        else:
            do_cores = [x for x in cores if x in avail_cores]
            if cores != do_cores:
                do_cores = [x for x in cores if x not in avail_cores]
                raise ValueError("Cores listed not available in the OMX file: {}".format(do_cores))

        avail_idx = src.list_mappings()
        if mappings is None:
            do_idx = avail_idx
            if not avail_idx:
                do_idx = ["zero_base_index"]
        else:
            do_idx = [x for x in mappings if x in avail_idx]
            if mappings != do_idx:
                do_idx = [x for x in mappings if x not in avail_idx]
                raise ValueError("Mappings/indices listed not available in the OMX file: {}".format(do_idx))

        shp = src.shape()
        if shp[0] != shp[1]:
            raise ValueError("AequilibraE only supports square matrices")
        zones = shp[0]

        if robust:
            # Use reduce as we have to keep track of which generated names have been used (to avoid collisions)
            core_names = functools.reduce(
                lambda acc, n: acc + [robust_name(n, CORE_NAME_MAX_LENGTH, acc)], do_cores, []
            )
            idx_names = functools.reduce(lambda acc, n: acc + [robust_name(n, INDEX_NAME_MAX_LENGTH, acc)], do_idx, [])
        else:
            core_names = [x for x in do_cores]
            idx_names = [x for x in do_idx]

        self.create_empty(
            file_name=file_path, zones=zones, matrix_names=core_names, index_names=idx_names, compressed=compressed
        )

        # Copy all cores
        for ncore, core in zip(core_names, do_cores):
            self.matrix[ncore][:, :] = np.array(src[core])[:, :]
        self.matrices.flush()

        # copy all indices
        if avail_idx:
            for nidx, idx in zip(idx_names, do_idx):
                ix = np.array(list(src.mapping(idx).keys()))
                self.indices[nidx][:] = ix[:]
        else:
            self.index[:] = np.arange(zones)

        self.indices.flush()

    def __load__(self):
        # GET File version
        self.__version__ = np.memmap(self.file_path, dtype="uint8", offset=0, mode="r+", shape=1)[0]

        if self.__version__ != VERSION:
            raise ValueError("Matrix formats do not match")

        # If matrix is compressed or not
        self.compressed = np.memmap(self.file_path, dtype="uint8", offset=1, mode="r+", shape=1)[0]

        # number matrix cells if compressed
        matrix_cells = np.memmap(self.file_path, dtype="uint64", offset=2, mode="r+", shape=1)[0]

        # Zones
        self.zones = np.memmap(self.file_path, dtype="uint32", offset=10, mode="r+", shape=1)[0]

        # Matrix cores
        self.cores = np.memmap(self.file_path, dtype="uint8", offset=14, mode="r+", shape=1)[0]

        # Matrix indices
        self.num_indices = np.memmap(self.file_path, dtype="uint8", offset=15, mode="r+", shape=1)[0]

        # Data type
        data_class = np.memmap(self.file_path, dtype="uint8", offset=16, mode="r+", shape=1)[0]

        # Data size
        data_size = np.memmap(self.file_path, dtype="uint8", offset=17, mode="r+", shape=1)[0]

        if data_class == INT:
            if data_size == 1:
                self.dtype = np.int8
            elif data_size == 2:
                self.dtype = np.int16
            elif data_size == 4:
                self.dtype = np.int32
            elif data_size == 8:
                self.dtype = np.int64
            elif data_size == 16:
                self.dtype = np.int128

        if data_class == FLOAT:
            if data_size == 2:
                self.dtype = np.float16
            elif data_size == 4:
                self.dtype = np.float32
            elif data_size == 8:
                self.dtype = np.float64
            elif data_size == 16:
                self.dtype = np.float128

        # matrix name
        self.name = np.memmap(self.file_path, dtype="S" + str(MATRIX_NAME_MAX_LENGTH), offset=18, mode="r+", shape=1)[0]

        # matrix description
        offset = 18 + MATRIX_NAME_MAX_LENGTH
        self.description = np.memmap(
            self.file_path, dtype="S" + str(MATRIX_DESCRIPTION_MAX_LENGTH), offset=offset, mode="r+", shape=1
        )[0]

        # core names
        offset += MATRIX_DESCRIPTION_MAX_LENGTH
        self.names = list(
            np.memmap(self.file_path, dtype="S" + str(CORE_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.cores)
        )
        self.names = [x.decode("utf-8") for x in self.names]

        # Index names
        offset += CORE_NAME_MAX_LENGTH * self.cores
        self.index_names = list(
            np.memmap(
                self.file_path, dtype="S" + str(INDEX_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.num_indices
            )
        )
        self.index_names = [x.decode("utf-8") for x in self.index_names]

        # Index
        offset += self.num_indices * INDEX_NAME_MAX_LENGTH
        self.indices = np.memmap(
            self.file_path, dtype="uint64", offset=offset, mode="r+", shape=(self.zones, self.num_indices)
        )
        self.set_index(self.index_names[0])

        # DATA
        offset += self.zones * 8 * self.num_indices
        if self.compressed:
            self.matrices = np.memmap(
                self.file_path, dtype=self.dtype, offset=offset, mode="r+", shape=(matrix_cells, self.cores + 2)
            )
        else:
            self.matrices = np.memmap(
                self.file_path, dtype=self.dtype, offset=offset, mode="r+", shape=(self.zones, self.zones, self.cores)
            )

        self.matrix = {}
        for i, v in enumerate(self.names):
            self.matrix[v] = self.matrices[:, :, i]
        self.matrix_hash = self.__builds_hash__()

    def __write__(self):
        np.memmap(self.file_path, dtype="uint8", offset=0, mode="w+", shape=1)[0] = self.__version__

        # If matrix is compressed or not
        np.memmap(self.file_path, dtype="uint8", offset=1, mode="r+", shape=1)[0] = self.compressed

        # number matrix cells if compressed
        matrix_cells = self.zones * self.zones
        np.memmap(self.file_path, dtype="uint64", offset=2, mode="r+", shape=1)[0] = matrix_cells

        # Zones
        np.memmap(self.file_path, dtype="uint32", offset=10, mode="r+", shape=1)[0] = self.zones

        # Matrix cores
        np.memmap(self.file_path, dtype="uint8", offset=14, mode="r+", shape=1)[0] = self.cores

        # Matrix indices
        np.memmap(self.file_path, dtype="uint8", offset=15, mode="r+", shape=1)[0] = self.num_indices

        # Data type
        data_class = self.define_data_class()
        np.memmap(self.file_path, dtype="uint8", offset=16, mode="r+", shape=1)[0] = data_class

        # Data size
        data_size = np.dtype(self.dtype).itemsize
        np.memmap(self.file_path, dtype="uint8", offset=17, mode="r+", shape=1)[0] = data_size

        # matrix name
        np.memmap(self.file_path, dtype="S" + str(MATRIX_NAME_MAX_LENGTH), offset=18, mode="r+", shape=1)[0] = self.name

        # matrix description
        offset = 18 + MATRIX_NAME_MAX_LENGTH
        np.memmap(self.file_path, dtype="S" + str(MATRIX_DESCRIPTION_MAX_LENGTH), offset=offset, mode="r+", shape=1)[
            0
        ] = self.description

        # core names
        offset += MATRIX_DESCRIPTION_MAX_LENGTH
        fp = np.memmap(
            self.file_path, dtype="S" + str(CORE_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.cores
        )
        for i, v in enumerate(self.names):
            fp[i] = v
        fp.flush()
        del fp

        # Index names
        offset += CORE_NAME_MAX_LENGTH * self.cores
        fp = np.memmap(
            self.file_path, dtype="S" + str(INDEX_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.num_indices
        )
        for i, v in enumerate(self.index_names):
            fp[i] = v
        fp.flush()
        del fp

        # Index
        offset += self.num_indices * INDEX_NAME_MAX_LENGTH
        self.indices = np.memmap(
            self.file_path, dtype="uint64", offset=offset, mode="r+", shape=(self.zones, self.num_indices)
        )
        self.indices.fill(0)
        self.indices.flush()
        self.set_index(self.index_names[0])

        offset += self.zones * 8 * self.num_indices
        if self.compressed:
            self.matrices = np.memmap(
                self.file_path, dtype=self.dtype, offset=offset, mode="r+", shape=(matrix_cells, self.cores + 2)
            )
        else:
            self.matrices = np.memmap(
                self.file_path, dtype=self.dtype, offset=offset, mode="r+", shape=(self.zones, self.zones, self.cores)
            )

        if np.issubdtype(self.dtype, np.integer):
            self.matrices.fill(np.iinfo(self.dtype).min)
        else:
            self.matrices.fill(np.nan)
        self.matrices.flush()

        self.matrix = {}
        for i, v in enumerate(self.names):
            self.matrix[v] = self.matrices[:, :, i]

    def set_index(self, index_to_set: str):
        """
                Sets the standard index to be the one the user wants to have be the one being used in all operations
                during run time. The first index is ALWAYS the default one every time the matrix is instantiated

                Parameters
                ----------
                index_to_set: string
                    Name of the index to be used. The default index name is 'main_index'

                ------------------------------------------------------------
                Example

                >>> zones_in_the_model = 3317
                >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']
                >>> index_list = ['tazs',  'census']

                >>> mat = AequilibraeMatrix()
                >>> mat.create_empty(file_name='my/path/to/file', zones=zones_in_the_model, matrix_names=names_list, index_names =index_list )
                >>> mat.num_indices
                2
                >>> mat.current_index
                'tazs'
                >>> mat.set_index('census')
                >>> mat.current_index
                'census'
                """

        if index_to_set in self.index_names:
            ind_index = self.index_names.index(index_to_set)
            self.index = self.indices[:, ind_index]
            self.current_index = index_to_set
        else:
            raise ValueError("Index {} needs to be a string or its integer index.".format(str(index_to_set)))

    def __getattr__(self, mat_name):
        if mat_name in object.__dict__:
            return self.__dict__[mat_name]

        if mat_name in self.names:
            return self.matrix[mat_name]

        raise AttributeError("No such method or matrix core! --> " + str(mat_name))

    # Transforms matrix from dense to CSR
    # def compress(self):
    #     print(self.file_path + '. Method not implemented yet. All matrices are dense')

    # Transforms matrix from CSR to dense
    # def decompress(self):
    #     print(self.file_path + '. Method not implemented yet. All matrices are dense')

    # Adds index to matrix
    # def add_index(self):
    #     print(self.file_path + '. Method not implemented yet. All indices need to exist during the matrix creation')

    # Adds index to matrix
    # def remove_index(self, index_number):
    #     print(self.file_path + '. Method not implemented. Indices are fixed on matrix creation: ' + str(index_number))

    def close(self):
        """
        Removes matrix from memory and flushes all data to disk
        """

        self.matrices.flush()
        self.index.flush()
        del self.matrices
        del self.index

    def export(self, output_name: str, cores: List[str] = None):
        """
        Exports the matrix to other formats. Formats currently supported: CSV, OMX

        When exporting to AEM or OMX, the user can chose to export only a set of cores, but all indices are exported

        When exporting to CSV, the active index will be used, and all cores will be exported as separate columns in
        the output file

        Parameters
        ----------
        output_name: Name of the output file

        cores: Names of the cores to be exported.

         ------------------------------------------------------------
        Example

        >>> zones_in_the_model = 3317
        >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

        >>> mat = AequilibraeMatrix()
        >>> mat.create_empty(file_name='my/path/to/file', zones=zones_in_the_model, matrix_names= names_list)
        >>> mat.cores
        ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

        >>> mat.export('my_new_path', ['Car trips', 'bike trips'])

        >>> mat2 = AequilibraeMatrix()
        >>> mat2.load('my_new_path')
        >>> mat2.cores
        ['Car trips', 'bike trips']
        """
        fname, file_extension = os.path.splitext(output_name.upper())

        if file_extension == ".OMX":
            spec = iutil.find_spec("openmatrix")
            if spec is None:
                raise ValueError("Open Matrix is not installed. Cannot continue")
            import openmatrix as omx

        if file_extension not in [".AEM", ".CSV", ".OMX"]:
            raise ValueError("File extension {} not implemented yet".format(file_extension))

        if cores is None:
            cores = self.names

        if file_extension == ".AEM":
            self.copy(output_name=output_name, cores=cores)

        elif file_extension == ".OMX":
            omx_export = omx.open_file(output_name, "w")
            for c in cores:
                omx_export[c] = self.matrix[c]

            for i, idx in enumerate(self.index_names):
                omx_export.create_mapping(idx, self.indices[:, i])
            omx_export.close()

        elif file_extension == ".CSV":
            names = self.view_names
            self.computational_view(cores)
            output = open(output_name, "w")

            titles = ["row", "column"]
            for core in self.view_names:
                titles.append(core)
            output.write(",".join(titles))

            for i in range(self.zones):
                for j in range(self.zones):
                    record = [self.index[i], self.index[j]]
                    if len(self.view_names) > 1:
                        record.extend(self.matrix_view[i, j, :])
                    else:
                        record.append(self.matrix_view[i, j])
                    output.write(",".join(str(x) for x in record))
            output.flush()
            output.close()
            self.computational_view(names)

    def load(self, file_path: str):
        """
                Loads matrix from disk. All cores and indices are load. First index is default

                Parameters
                ----------
                file_path: Path to AEM file on disk
                ------------------------------------------------------------
                Example

                >>> zones_in_the_model = 3317
                >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

                >>> mat = AequilibraeMatrix()
                >>> mat.create_empty(file_name='my/path/to/file', zones=zones_in_the_model, matrix_names= names_list)
                >>> mat.close()

                >>> mat2 = AequilibraeMatrix()
                >>> mat2.load('my/path/to/file')
                >>> mat2.zones
                3317
                """

        self.file_path = file_path
        self.__load__()

    def computational_view(self, core_list: List[str] = None):
        """
        Creates a memory view for a list of matrices that is compatible with Cython memory buffers

        It allows for AequilibraE matrices to be used in all parallelized algorithms within AequilibraE

        Parameters
        ----------
        core_list: List with the names of all matrices that need to be in the buffer
        ------------------------------------------------------------
        Example

        >>> zones_in_the_model = 3317
        >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

        >>> mat = AequilibraeMatrix()
        >>> mat.create_empty(file_name='my/path/to/file', zones=zones_in_the_model, matrix_names= names_list)
        >>> mat.computational_view(['bike trips', 'walk trips'])
        >>> mat.view_names
        ['bike trips', 'walk trips']
        """

        self.matrix_view = None
        self.view_names = None
        if core_list is None:
            core_list = self.names
        else:
            if isinstance(core_list, list):
                for i in core_list:
                    if i not in self.names:
                        raise ValueError("Matrix core {} no available on this matrix".format(i))

                if len(core_list) > 1:
                    for i, x in enumerate(core_list[1:]):
                        k = self.names.index(x)  # index of the first element
                        k0 = self.names.index(core_list[i])  # index of the first element
                        if k - k0 != 1:
                            raise ValueError("Matrix cores {} and {} are not adjacent".format(core_list[i - 1], x))
            else:
                raise TypeError("Please provide a list of matrices")

        self.view_names = core_list
        if len(core_list) == 1:
            # self.matrix_view = self.matrix[:, :, self.names.index(core_list[0]):self.names.index(core_list[0])+1]
            self.matrix_view = self.matrices[:, :, self.names.index(core_list[0])]
        elif len(core_list) > 1:
            self.matrix_view = self.matrices[:, :, self.names.index(core_list[0]) : self.names.index(core_list[-1]) + 1]

    def copy(self, output_name: str = None, cores: List[str] = None, names: List[str] = None, compress: bool = None):
        """
        Copies a list of cores (or all cores) from one matrix file to another one

        Parameters
        ----------
        output_name: Name of the new matrix file

        cores: List (str)
            List of the matrix cores to be copied

        names: List(str), optional
            List with the new names for the cores (same list length as cores)

        compress: bool
            Whether you want to compress the matrix or not. NOT YET IMPLEMENTED
        ------------------------------------------------------------
        Example

        >>> zones_in_the_model = 3317
        >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

        >>> mat = AequilibraeMatrix()
        >>> mat.create_empty(file_name='my/path/to/file', zones=zones_in_the_model, matrix_names= names_list)
        >>> mat.copy('my/new/path/to/file', cores=['bike trips', 'walk trips'], names=['bicycle', 'walking'])

        >>> mat2 = AequilibraeMatrix()
        >>> mat2.load('my/new/path/to/file')
        >>> mat.cores
        ['bicycle', 'walking']
        """
        if output_name is None:
            output_name = self.random_name()

        if cores is None:
            copyfile(self.file_path, output_name)
            output = AequilibraeMatrix()
            output.load(output_name)
            if self.view_names is not None:
                output.computational_view(self.view_names)

            if compress is not None:
                if compress != self.compressed:
                    if compress:
                        output.compress()
                    else:
                        output.decompress()
        else:
            if compress is None:
                compress = self.compressed

            if not isinstance(cores, list):
                raise ValueError("Cores need to be presented as list")

            for i in cores:
                if i not in self.names:
                    raise ValueError("Matrix core {} not available on this matrix".format(i))

            if names is None:
                names = cores
            else:
                if not isinstance(names, list):
                    raise ValueError("Names need to be presented as list")

                if len(names) != len(cores):
                    raise ValueError("Number of cores to cpy and list of names are not compatible")

            output = AequilibraeMatrix()
            output.create_empty(
                file_name=output_name,
                zones=self.zones,
                matrix_names=names,
                index_names=self.index_names,
                data_type=self.dtype,
                compressed=bool(compress),
            )

            output.indices[:] = self.indices[:]
            for i, c in enumerate(cores):
                output.matrices[:, :, i] = self.matrices[:, :, self.names.index(c)]
            self.matrix_hash = output.__builds_hash__()
            output.matrices.flush()

        return output

    def rows(self):
        # type: () -> np.array()
        """
            Returns row vector for the matrix in the computational view

            Computational view needs to be set to a single matrix core

            ------------------------------------------------------------
            Example

            >>> mat = AequilibraeMatrix()
            >>> mat.load('my/path/to/file')
            >>> mat.computational_view(mat.cores[0])
            >>> mat.rows()
            array([0.,...,0.])
            """
        return self.__vector(axis=0)

    def columns(self):
        # type: () -> np.array()
        """
        Returns column vector for the matrix in the computational view

        Computational view needs to be set to a single matrix core

        ------------------------------------------------------------
        Example

        >>> mat = AequilibraeMatrix()
        >>> mat.load('my/path/to/file')
        >>> mat.computational_view(mat.cores[0])
        >>> mat.columns()
        array([0.,...,0.])
        """
        return self.__vector(axis=1)

    def nan_to_num(self):
        """
        Converts all NaN values in all cores in the computational view to zeros

        ------------------------------------------------------------
        Example

        >>> mat = AequilibraeMatrix()
        >>> mat.load('my/path/to/file')
        >>> mat.computational_view(mat.cores[0])
        >>> mat.nan_to_num()
        """
        if np.issubdtype(self.dtype, np.floating) and self.matrix_view is not None:
            for m in self.view_names:
                self.matrix[m][:, :] = np.nan_to_num(self.matrix[m])[:, :]

    def __vector(self, axis: int):
        if self.view_names is None:
            raise ReferenceError("Matrix is not set for computation")
        if len(self.view_names) > 1:
            raise ValueError("Vector for a multi-core matrix is ambiguous")
        return self.matrix_view.astype(np.float).sum(axis=axis)[:]

    def __builds_hash__(self):
        return {self.index[i]: i for i in range(self.zones)}

    def define_data_class(self):
        if np.issubdtype(self.dtype, np.floating):
            data_class = FLOAT
        elif np.issubdtype(self.dtype, np.integer):
            data_class = INT
        else:
            raise ValueError(
                "Data type not supported. You can choose Integers of 8, 16, 32 and 64 bits, "
                "or floats with 16, 32 or 64 bits"
            )
        return data_class

    def setName(self, matrix_name: str):
        """
        Sets the name for the matrix itself

        Parameters
        ----------
        matrix_name: str
            matrix name. Maximum length is 50 characters

        ------------------------------------------------------------
        Example


        >>> mat = AequilibraeMatrix()
        >>> mat.load('my/path/to/file')
        >>> mat.setName('This is my example')
        >>> mat.name
        'This is my example'
        """

        if matrix_name is not None:
            if len(str(matrix_name)) > MATRIX_NAME_MAX_LENGTH:
                matrix_name = str(matrix_name)[0:MATRIX_NAME_MAX_LENGTH]

            np.memmap(self.file_path, dtype="S" + str(MATRIX_NAME_MAX_LENGTH), offset=18, mode="r+", shape=1)[
                0
            ] = matrix_name

    def setDescription(self, matrix_description: str):
        """
        Sets description for the matrix

        Parameters
        ----------
        matrix_name: str
            Text with matrix description . Maximum length is 144 characters

        ------------------------------------------------------------
        Example


        >>> mat = AequilibraeMatrix()
        >>> mat.load('my/path/to/file')
        >>> mat.setDescription('This is some text about this matrix of mine')
        >>> mat.description
        'This is some text about this matrix of mine'
        """
        if matrix_description is not None:
            if len(str(matrix_description)) > MATRIX_DESCRIPTION_MAX_LENGTH:
                matrix_description = str(matrix_description)[0:MATRIX_DESCRIPTION_MAX_LENGTH]

            np.memmap(
                self.file_path,
                dtype="S" + str(MATRIX_NAME_MAX_LENGTH),
                offset=18 + MATRIX_NAME_MAX_LENGTH,
                mode="r+",
                shape=1,
            )[0] = matrix_description

    @staticmethod
    def random_name() -> str:
        """
        Returns a random name for a matrix with root in the temp directory of the user

        ------------------------------------------------------------
        Example


        >>> name = AequilibraeMatrix().random_name()
        '/tmp/Aequilibrae_matrix_54625f36-bf41-4c85-80fb-7fc2e3f3d76e.aem'
        """
        return os.path.join(tempfile.gettempdir(), "Aequilibrae_matrix_" + str(uuid.uuid4()) + ".aem")



from aequilibrae.starts_logging import StartsLogging
from aequilibrae import distribution
from aequilibrae import matrix
from aequilibrae import paths
from aequilibrae import transit
from aequilibrae.paths import release_version as __version__

name = "aequilibrae"

StartsLogging()



import os
from tempfile import gettempdir
from unittest import TestCase

from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths import allOrNothing
from aequilibrae.paths import Graph
from aequilibrae.paths import AssignmentResults
from ...data import test_graph


# TODO: Add checks for results for this test (Assignment AoN)
class TestAllOrNothing(TestCase):
    def test_execute(self):
        # Loads and prepares the graph
        g = Graph()
        g.load_from_disk(test_graph)
        g.set_graph(cost_field="distance", skim_fields=None)
        # None implies that only the cost field will be skimmed

        # Prepares the matrix for assignment
        args = {
            "file_name": os.path.join(gettempdir(), "my_matrix.aem"),
            "zones": g.num_zones,
            "matrix_names": ["cars", "trucks"],
            "index_names": ["my indices"],
        }

        matrix = AequilibraeMatrix()
        matrix.create_empty(**args)

        matrix.index[:] = g.centroids[:]
        matrix.cars.fill(1)
        matrix.trucks.fill(2)
        matrix.computational_view(["cars"])

        # Performs assignment
        res = AssignmentResults()
        res.prepare(g, matrix)

        assig = allOrNothing(matrix, g, res)
        assig.execute()

        res.save_to_disk(os.path.join(gettempdir(), "link_loads.aed"))
        res.save_to_disk(os.path.join(gettempdir(), "link_loads.csv"))

        matrix.computational_view()
        # Performs assignment
        res = AssignmentResults()
        res.prepare(g, matrix)

        assig = allOrNothing(matrix, g, res)
        assig.execute()
        res.save_to_disk(os.path.join(gettempdir(), "link_loads_2_classes.aed"))
        res.save_to_disk(os.path.join(gettempdir(), "link_loads_2_classes.csv"))



import os
import sys
import unittest
from aequilibrae.paths import Graph
from aequilibrae.paths.results import SkimResults
from aequilibrae.paths import skimming_single_origin
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
import numpy as np
from .parameters_test import centroids

# Adds the folder with the data to the path and collects the paths to the files
# lib_path = os.path.abspath(os.path.join('..', '../tests'))
# sys.path.append(lib_path)
from ...data import path_test, test_graph


class TestSkimming_single_origin(unittest.TestCase):
    def test_skimming_single_origin(self):

        origin = 1

        # graph
        g = Graph()
        g.load_from_disk(test_graph)
        g.set_graph(cost_field="distance", skim_fields=None)
        # g.block_centroid_flows = False
        # None implies that only the cost field will be skimmed

        # skimming results
        res = SkimResults()
        res.prepare(g)
        aux_result = MultiThreadedNetworkSkimming()
        aux_result.prepare(g, res)

        a = skimming_single_origin(origin, g, res, aux_result, 0)
        tot = np.sum(res.skims.distance[origin, :])
        if tot > 10e10:
            self.fail("Skimming was not successful. At least one np.inf returned.")

        if a != origin:
            self.fail("Skimming returned an error: " + a)



import logging
import pickle
import uuid
from datetime import datetime

import numpy as np

from .__version__ import binary_version as VERSION

""""""
"""-----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Transportation graph class
 Purpose:    Implement a standard graph class to support all network computation

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    05/June/2015
 Updated:    03/Dec/2017
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------"""


class Graph(object):
    """
    Graph class
    """

    def __init__(self):
        self.logger = logging.getLogger("aequilibrae")
        self.__integer_type = np.int64
        self.__float_type = np.float64

        self.required_default_fields = []
        self.reset_single_fields()
        self.other_fields = ""
        self.date = str(datetime.now())

        self.description = "No description added so far"

        self.num_links = -1
        self.num_nodes = -1
        self.num_zones = -1
        self.network = False  # This method will hold ALL information on the network
        self.graph = False  # This method will hold an array with ALL fields in the graph.

        # These are the fields actually used in computing paths
        self.all_nodes = False  # Holds an array with all nodes in the original network
        self.nodes_to_indices = False  # Holds the reverse of the all_nodes
        self.fs = False  # This method will hold the forward star for the graph
        self.b_node = False  # b node for each directed link

        self.cost = None  # This array holds the values being used in the shortest path routine
        self.skims = False  # 2-D Array with the fields to be computed as skims
        self.skim_fields = False  # List of skim fields to be used in computation
        self.cost_field = False  # Name of the cost field
        self.ids = False  # 1-D Array with link IDs (sequence from 0 to N-1)

        self.block_centroid_flows = False
        self.penalty_through_centroids = np.inf

        self.centroids = None  # NumPy array of centroid IDs

        self.status = "NO network loaded"
        self.network_ok = False
        self.type_loaded = False
        self.__version__ = VERSION

        # Randomly generate a unique Graph ID randomly
        self.__id__ = uuid.uuid4().hex
        self.__source__ = None  # Name of the file that originated the graph

        # In case the graph is generated in QGIS, it is useful to have the name of the layer and fields that originated
        # it
        self.__field_name__ = None
        self.__layer_name__ = None

    def default_types(self, tp: str):
        if tp == "int":
            return self.__integer_type
        elif tp == "float":
            return self.__float_type
        else:
            raise ValueError("It must be either a int or a float")

    # Create a graph from a shapefile. To be upgraded to ANY geographic file in the future
    def create_from_geography(
        self,
        geo_file: str,
        id_field: str,
        dir_field: str,
        cost_field: str,
        centroids: np.ndarray,
        skim_fields=[],
        anode="A_NODE",
        bnode="B_NODE",
    ) -> None:
        """
        :param geo_file: Path to the geographic file to be used. File is usually the output of the network preparation
                         tool from the AequilibraE plugin for QGIS
        :param id_field: Name of the field that has link IDs (must be unique)
        :param dir_field: Name of the field that has the link directions of flow ([-1, 0, 1])
        :param cost_field: Name of the field that has the cost data (field to minimized in shortest para)
        :param centroids: Numpy Array with a list of centroids included in this graph
        :param skim_fields: List with the name of fields to be skimmed
        :param anode: Name of the field with information of A_Node of links (if different than *A_NODE*)
        :param bnode: Name of the field with information of B_Node of links (if different than *B_NODE*)
        """

        import shapefile

        cost_field_name = cost_field
        error = None
        geo_file_records = shapefile.Reader(geo_file)
        records = geo_file_records.records()

        def find_field_index(fields, field_name):
            for i, f in enumerate(fields):
                if f[0] == field_name:
                    return i - 1

            f = [str(x[0]) for x in fields]
            raise ValueError(field_name + " does not exist. Fields available are: " + ", ".join(f))
            return -1

        # collect the fields in the network
        check_titles = [id_field, dir_field, anode, bnode, cost_field]
        id_field = find_field_index(geo_file_records.fields, id_field)
        dir_field = find_field_index(geo_file_records.fields, dir_field)
        cost_field = find_field_index(geo_file_records.fields, cost_field)
        anode = find_field_index(geo_file_records.fields, anode)
        bnode = find_field_index(geo_file_records.fields, bnode)

        # Appends all fields to the list of fields to be used
        all_types = [
            self.__integer_type,
            self.__integer_type,
            self.__integer_type,
            self.__float_type,
            self.__float_type,
            np.int8,
        ]
        all_titles = [
            "link_id",
            "a_node",
            "b_node",
            cost_field_name.lower() + "_ab",
            cost_field_name.lower() + "_ba",
            "direction",
        ]
        check_fields = [id_field, dir_field, anode, bnode, cost_field]
        types_to_check = [int, int, int, int, float]

        # Loads the skim index fields
        dict_field = {}
        for k in skim_fields:
            skim_index = find_field_index(geo_file_records.fields, k)
            check_fields.append(skim_index)
            check_titles.append(k)
            types_to_check.append(float)

            all_types.append(self.__float_type)
            all_types.append(self.__float_type)
            all_titles.append((k + "_ab"))
            all_titles.append((k + "_ba"))
            dict_field[k + "_ab"] = skim_index
            dict_field[k + "_ba"] = skim_index

        dt = [(t, d) for t, d in zip(all_titles, all_types)]

        # Check ID uniqueness and if there are any non-valid values
        all_ids = []
        for feat in records:
            for i, j in enumerate(check_fields):
                k = feat[j]
                if not isinstance(k, types_to_check[i]):
                    error = check_titles[i], "field has wrong type or empty values"
                    break
            all_ids.append(feat[check_fields[0]])
            if error is not None:
                break

        if error is None:
            # Checking uniqueness
            all_ids = np.array(all_ids, np.int)
            y = np.bincount(all_ids)
            if np.max(y) > 1:
                error = "IDs are not unique."

        if error is None:
            data = []

            for feat in records:
                line = []
                line.append(feat[id_field])
                line.append(feat[anode])
                line.append(feat[bnode])
                line.append(feat[cost_field])
                line.append(feat[cost_field])
                line.append(feat[dir_field])

                # We append the skims now
                for k in all_titles:
                    if k in dict_field:
                        line.append(feat[dict_field[k]])
                data.append(line)

            network = np.asarray(data)
            del data

            self.network = np.zeros(network.shape[0], dtype=dt)
            for k, t in enumerate(dt):
                self.network[t[0]] = network[:, k].astype(t[1])
            del network

            self.type_loaded = "SHAPEFILE"
            self.status = "OK"
            self.network_ok = True
            self.prepare_graph(centroids)
            self.__source__ = geo_file
            self.__field_name__ = None
            self.__layer_name__ = None
        if error is not None:
            raise ValueError(error)

    def prepare_graph(self, centroids: np.ndarray) -> None:
        """
        :param centroids: Array with centroid IDs. Mandatory type Int64, unique and positive
        """

        # Creates the centroids
        if centroids is not None and isinstance(centroids, np.ndarray):
            if np.issubdtype(centroids.dtype, np.integer):
                if centroids.min() <= 0:
                    raise ValueError("Centroid IDs need to be positive")
                else:
                    if centroids.shape[0] != np.unique(centroids).shape[0]:
                        raise ValueError("Centroid IDs are not unique")
                self.centroids = centroids
            else:
                raise ValueError("Centroids need to be an array of integers 64 bits")
        else:
            raise ValueError("Centroids need to be a NumPy array of integers 64 bits")
        self.build_derived_properties()

        if not self.network_ok:
            raise ValueError("Network not yet properly loaded")
        else:
            all_titles = self.network.dtype.names

            if self.status == "OK":
                negs = self.network[self.network["direction"] == -1]
                poss = self.network[self.network["direction"] == 1]
                zers = self.network[self.network["direction"] == 0]

                self.num_links = negs.shape[0] + poss.shape[0]
                self.num_links += zers.shape[0] * 2

                dtype = self.__build_dtype(all_titles)

                self.graph = np.zeros(self.num_links, dtype=dtype)
                a1 = negs.shape[0]
                a2 = a1 + poss.shape[0]
                a3 = a2 + zers.shape[0]
                a4 = a3 + zers.shape[0]

                # Create the graph-specific node numbers
                self.all_nodes = np.unique(np.hstack((self.network["a_node"], self.network["b_node"]))).astype(
                    self.__integer_type
                )
                # We put the centroids as the first N elements of this array
                for i in self.centroids:
                    self.all_nodes = np.delete(self.all_nodes, np.argwhere(self.all_nodes == i))

                self.all_nodes = np.hstack((centroids, self.all_nodes)).astype(self.__integer_type)
                self.num_nodes = self.all_nodes.shape[0]
                self.nodes_to_indices = np.empty(int(self.all_nodes.max()) + 1, self.__integer_type)
                self.nodes_to_indices.fill(-1)
                self.nodes_to_indices[self.all_nodes] = np.arange(self.num_nodes)

                for i in all_titles:
                    if i == "link_id":
                        self.graph[i][0:a1] = negs[i]
                        self.graph[i][a1:a2] = poss[i]
                        self.graph[i][a2:a3] = zers[i]
                        self.graph[i][a3:a4] = zers[i]

                    elif i == "a_node":
                        self.graph[i][0:a1] = self.nodes_to_indices[negs["b_node"]]
                        self.graph[i][a1:a2] = self.nodes_to_indices[poss[i]]
                        self.graph[i][a2:a3] = self.nodes_to_indices[zers["b_node"]]
                        self.graph[i][a3:a4] = self.nodes_to_indices[zers[i]]

                    elif i == "b_node":
                        self.graph[i][0:a1] = self.nodes_to_indices[negs["a_node"]]
                        self.graph[i][a1:a2] = self.nodes_to_indices[poss[i]]
                        self.graph[i][a2:a3] = self.nodes_to_indices[zers["a_node"]]
                        self.graph[i][a3:a4] = self.nodes_to_indices[zers[i]]

                    elif i == "direction":
                        self.graph[i][0:a1] = -1
                        self.graph[i][a1:a2] = 1
                        self.graph[i][a2:a3] = -1
                        self.graph[i][a3:a4] = 1
                    else:
                        if i[-3:] == "_ab":
                            self.graph[i[0:-3]][0:a1] = negs[i[0:-3] + "_ba"]
                            self.graph[i[0:-3]][a1:a2] = poss[i]
                            self.graph[i[0:-3]][a2:a3] = zers[i[0:-3] + "_ba"]
                            self.graph[i[0:-3]][a3:a4] = zers[i]

                ind = np.lexsort((self.graph["b_node"], self.graph["a_node"]))
                self.graph = self.graph[ind]
                del ind
                self.graph["id"] = np.arange(self.num_links)
                self.fs = np.zeros(self.num_nodes + 1, dtype=self.__integer_type)

                a = self.graph["a_node"][0]
                p = 0
                k = 0
                for i in range(1, self.num_links):
                    if a != self.graph["a_node"][i]:
                        for j in range(p, self.graph["a_node"][i]):
                            self.fs[j + 1] = k
                        p = a
                        a = self.graph["a_node"][i]
                        k = i

                for j in range(p, self.num_nodes):
                    self.fs[j + 1] = k

                self.fs[self.num_nodes] = self.graph.shape[0]
                self.ids = self.graph["id"]
                self.b_node = np.array(self.graph["b_node"], self.__integer_type)

    def __build_dtype(self, all_titles):
        dtype = [
            ("link_id", self.__integer_type),
            ("a_node", self.__integer_type),
            ("b_node", self.__integer_type),
            ("direction", np.int8),
            ("id", self.__integer_type),
        ]
        for i in all_titles:
            if i not in self.required_default_fields and i[0:-3] not in self.required_default_fields:
                if i[-3:] != "_ab":
                    dtype.append((i[0:-3], self.network[i].dtype))
        return dtype

    # We set which are the fields that are going to be minimized in this file
    # TODO: Change the call for all the uses on this function
    def set_graph(self, cost_field=None, skim_fields=False, block_centroid_flows=None):
        """
        :type cost_field
        :type block_centroid_flows
        :type skim_fields: list of fields for skims
        :type self: object
        """
        if block_centroid_flows is not None:
            if isinstance(block_centroid_flows, bool):
                self.set_blocked_centroid_flows(block_centroid_flows)

            else:
                raise ValueError("block_c" "entroid_flows needs to be a boolean")

        if cost_field is not None:
            if cost_field in self.graph.dtype.names:
                self.cost_field = cost_field
                if self.graph[cost_field].dtype == self.__float_type:
                    self.cost = self.graph[cost_field]
                else:
                    self.cost = self.graph[cost_field].astype(self.__float_type)
                    Warning("Cost field with wrong type. Converting to float64")

            else:
                raise ValueError("cost_field not available in the graph:" + str(self.graph.dtype.names))

        if self.cost_field is not None:
            if not skim_fields:
                skim_fields = [self.cost_field]
            else:
                s = [self.cost_field]
                for i in skim_fields:
                    if i in self.graph.dtype.names:
                        if i not in s:
                            s.append(i)
                    else:
                        self.skim_fields = None
                        self.skims = None
                        raise ValueError("Skim", i, " not available in the graph:", self.graph.dtype.names)
                skim_fields = s
        else:
            if skim_fields:
                raise ValueError("Before setting skims, you need to set the cost field")

        t = False
        for i in skim_fields:
            if self.graph[i].dtype != self.__float_type:
                t = True

        self.skims = np.zeros((self.num_links, len(skim_fields) + 1), self.__float_type)

        if t:
            Warning("Some skim field with wrong type. Converting to float64")
            for i, j in enumerate(skim_fields):
                self.skims[:, i] = self.graph[j].astype(self.__float_type)
        else:
            for i, j in enumerate(skim_fields):
                self.skims[:, i] = self.graph[j]
        self.skim_fields = skim_fields

        self.build_derived_properties()
        return True

    def set_blocked_centroid_flows(self, blocking):
        if self.num_zones > 0:
            self.block_centroid_flows = blocking
            self.b_node = np.array(self.graph["b_node"], self.__integer_type)
        else:
            raise ValueError("You can only block flows through centroids after setting the centroids")

    # Procedure to pickle graph and save to disk
    def save_to_disk(self, filename):
        mygraph = {}
        mygraph["description"] = self.description
        mygraph["num_links"] = self.num_links
        mygraph["num_nodes"] = self.num_nodes
        mygraph["network"] = self.network
        mygraph["graph"] = self.graph
        mygraph["all_nodes"] = self.all_nodes
        mygraph["nodes_to_indices"] = self.nodes_to_indices
        mygraph["num_nodes"] = self.num_nodes
        mygraph["fs"] = self.fs
        mygraph["b_node"] = self.b_node
        mygraph["cost"] = self.cost
        mygraph["cost_field"] = self.cost_field
        mygraph["skims"] = self.skims
        mygraph["skim_fields"] = self.skim_fields
        mygraph["ids"] = self.ids
        mygraph["block_centroid_flows"] = self.block_centroid_flows
        mygraph["centroids"] = self.centroids
        mygraph["status"] = self.status
        mygraph["network_ok"] = self.network_ok
        mygraph["type_loaded"] = self.type_loaded
        mygraph["graph_id"] = self.__id__
        mygraph["graph_version"] = self.__version__

        with open(filename, "wb") as f:
            pickle.dump(mygraph, f)

    def load_from_disk(self, filename):
        with open(filename, "rb") as f:
            mygraph = pickle.load(f)
            self.description = mygraph["description"]
            self.num_links = mygraph["num_links"]
            self.num_nodes = mygraph["num_nodes"]
            self.network = mygraph["network"]
            self.graph = mygraph["graph"]
            self.all_nodes = mygraph["all_nodes"]
            self.nodes_to_indices = mygraph["nodes_to_indices"]
            self.num_nodes = mygraph["num_nodes"]
            self.fs = mygraph["fs"]
            self.b_node = mygraph["b_node"]
            self.cost = mygraph["cost"]
            self.cost_field = mygraph["cost_field"]
            self.skims = mygraph["skims"]
            self.skim_fields = mygraph["skim_fields"]
            self.ids = mygraph["ids"]
            self.block_centroid_flows = mygraph["block_centroid_flows"]
            self.centroids = mygraph["centroids"]
            self.status = mygraph["status"]
            self.network_ok = mygraph["network_ok"]
            self.type_loaded = mygraph["type_loaded"]
            self.__id__ = mygraph["graph_id"]
            self.__version__ = mygraph["graph_version"]
        self.build_derived_properties()

    def build_derived_properties(self):
        if self.centroids is not None:
            self.num_zones = self.centroids.shape[0]

    # We return the list of the fields that are the same for both directions to their initial states
    def reset_single_fields(self):
        self.required_default_fields = ["link_id", "a_node", "b_node", "direction", "id"]

    # We add a new fields that is the same for both directions
    def add_single_field(self, new_field):
        if new_field not in self.required_default_fields:
            self.required_default_fields.append(new_field)

    def available_skims(self):
        graph_fields = list(self.graph.dtype.names)
        return [x for x in graph_fields if x not in ["link_id", "a_node", "b_node", "direction", "id"]]

    # We check if all minimum fields are there
    def __network_error_checking__(self):

        # Checking field names
        has_fields = self.network.dtype.names
        must_fields = ["link_id", "a_node", "b_node", "direction"]
        for field in must_fields:
            if field not in has_fields:
                self.status = 'could not find field "%s" in the network array' % field

                # checking data types
        must_types = [self.__integer_type, self.__integer_type, self.__integer_type, np.int8]
        for field, ytype in zip(must_fields, must_types):
            if self.network[field].dtype != ytype:
                self.status = (
                    'Field "%s" in the network array has the wrong type. Please refer to the documentation' % field
                )

                # Uniqueness of the id
        link_ids = self.network["link_id"].astype(np.int)
        if link_ids.shape[0] != np.unique(link_ids).shape[0]:
            self.status = '"link_id" field not unique'

            # Direction values
        if np.max(self.network["direction"]) > 1 or np.min(self.network["direction"]) < -1:
            self.status = '"direction" field not limited to (-1,0,1) values'

    # Needed for when we load the graph directly
    def __graph_error_checking__(self):
        # Checking field names
        self.status = "graph loaded"
        has_fields = self.graph.dtype.names
        must_fields = ["link_id", "a_node", "b_node", "id"]
        for field in must_fields:
            if field not in has_fields:
                self.status = 'could not find field "%s" in the network array' % field

                # checking data types
        must_types = [self.__integer_type, self.__integer_type, self.__integer_type, self.__integer_type]
        for field, ytype in zip(must_fields, must_types):
            if self.graph[field].dtype != ytype:
                self.status = (
                    'Field "%s" in the network array has the wrong type. Please refer to the documentation' % field
                )

                # Uniqueness of the graph id
        a = self.graph["id"].astype(np.int)
        if a.shape[0] != np.unique(a).shape[0]:
            self.status = '"id" field not unique'

        a = np.bincount(self.graph["link_id"].astype(np.int))
        if np.max(a) > 2:
            self.status = '"link_id" field has more than one link per direction'

        if np.min(self.graph["id"]) != 0:
            self.status = '"id" field needs to start in 0 and go to number of links - 1'

        if np.max(self.graph["id"]) > self.graph["id"].shape[0] - 1:
            self.status = '"id" field needs to start in 0 and go to number of links - 1'

    def __determine_types__(self, new_type, current_type):

        if new_type.isdigit():
            new_type = int(new_type)
        else:
            try:
                new_type = float(new_type)
            except ValueError as verr:
                self.logger.warning("Could not convert {} - {}".format(new_type, verr.__str__()))
        nt = type(new_type)
        def_type = None
        if nt == int:
            def_type = int
            if current_type == float:
                def_type == float
            elif current_type == str:
                def_type = str
        elif nt == float:
            def_type = float
            if current_type == str:
                def_type = str
        elif nt == str:
            def_type = str
        else:
            raise ValueError("WRONG TYPE OR NULL VALUE")
        return def_type



import numpy as np 

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

ITYPE = np.int64
ctypedef np.int64_t ITYPE_t

# EPS is the precision of DTYPE
cdef DTYPE_t DTYPE_EPS = 1E-15

# NULL_IDX is the index used in predecessor matrices to store a non-path
cdef ITYPE_t NULL_IDX = 18446744073709551615

cdef double INFINITE = 1.79769313e+308

VERSION = 0.5
MINOR_VRSN = 2
binary_version = "0.5.0"
release_name = "Karlsruhe"


import multiprocessing as mp
import numpy as np
import warnings
import sqlite3
import sys
import os
from numpy.lib.format import open_memmap
from ...matrix import AequilibraeMatrix, AequilibraEData
from ..graph import Graph

"""
TO-DO:
1. Create a file type for memory-mapped path files
   Same idea of the AequilibraEData container, but using the format.memmap from NumPy
2. Make the writing to SQL faster by disabling all checks before the actual writing
"""


class AssignmentResults:
    def __init__(self):
        """
        @type graph: Set of numpy arrays to store Computation results
        self.critical={required:{"links":[lnk_id1, lnk_id2, ..., lnk_idn], "path file": False}, results:{}}
        """
        self.link_loads = None  # The actual results for assignment
        self.skims = None  # The array of skims
        self.no_path = None  # The list os paths
        self.num_skims = None  # number of skims that will be computed. Depends on the setting of the graph provided
        self.cores = mp.cpu_count()
        self.classes = {"number": 1, "names": ["flow"]}

        self.critical_links = {"save": False, "queries": {}, "results": False}  # Queries are a dictionary

        self.link_extraction = {"save": False, "queries": {}, "output": None}  # Queries are a dictionary

        self.path_file = {"save": False, "results": None}

        self.nodes = -1
        self.zones = -1
        self.links = -1
        self.__graph_id__ = None
        self.__float_type = None
        self.__integer_type = None

        self.lids = None
        self.direcs = None

    # In case we want to do by hand, we can prepare each method individually
    def prepare(self, graph, matrix):
        """
        :param graph: AequilibraE graph. Needs to have been set with number of centroids and list of skims (if any)
        :type graph: Graph
        :param matrix: AequilibraE Matrix properly set for computation using matrix.computational_view([matrix list])
        :type matrix: AequilibraeMatrix
        :return:
        """

        self.__float_type = graph.default_types("float")
        self.__integer_type = graph.default_types("int")

        if matrix.view_names is None:
            raise ("Please set the matrix_procedures computational view")
        else:
            self.classes["number"] = 1
            if len(matrix.matrix_view.shape) > 2:
                self.classes["number"] = matrix.matrix_view.shape[2]
            self.classes["names"] = matrix.view_names

        if graph is None:
            raise ("Please provide a graph")
        else:
            self.nodes = graph.num_nodes
            self.zones = graph.num_zones
            self.centroids = graph.centroids
            self.links = graph.num_links
            self.num_skims = graph.skims.shape[1]
            self.skim_names = [x for x in graph.skim_fields]
            self.lids = graph.graph["link_id"]
            self.direcs = graph.graph["direction"]
            self.__redim()
            self.__graph_id__ = graph.__id__

            self.setSavePathFile(False)
            self.setCriticalLinks(False)

    def reset(self):
        if self.link_loads is not None:
            self.skims.matrices.fill(0)
            self.no_path.fill(0)
        else:
            raise ValueError("Exception: Assignment results object was not yet prepared/initialized")

    def __redim(self):
        self.link_loads = np.zeros((self.links, self.classes["number"]), self.__float_type)
        self.skims = np.zeros((self.zones, self.zones, self.num_skims), self.__float_type)
        self.skims = AequilibraeMatrix()

        self.skims.create_empty(file_name=self.skims.random_name(), zones=self.zones, matrix_names=self.skim_names)
        self.skims.index[:] = self.centroids[:]
        self.skims.computational_view()
        if len(self.skims.matrix_view.shape[:]) == 2:
            self.skims.matrix_view = self.skims.matrix_view.reshape((self.zones, self.zones, 1))
        self.no_path = np.zeros((self.zones, self.zones), dtype=self.__integer_type)

        self.reset()

    def set_cores(self, cores):
        if isinstance(cores, int):
            if cores > 0:
                if self.cores != cores:
                    self.cores = cores
                    if self.link_loads is not None:
                        self.__redim()
            else:
                raise ValueError("Number of cores needs to be equal or bigger than one")
        else:
            raise ValueError("Number of cores needs to be an integer")

    def setCriticalLinks(self, save=False, queries={}, crit_res_result=None):
        a = AequilibraeMatrix()
        if save:
            if crit_res_result is None:
                warnings.warn("Critical Link analysis not set properly. Need to specify output file too")
            else:
                if crit_res_result[-3:].lower() != "aem":
                    crit_res_result += ".aes"

                if self.nodes > 0 and self.zones > 0:
                    if ["elements", "labels", "type"] in queries.keys():
                        if len(queries["labels"]) == len(queries["elements"]) == len(queries["type"]):
                            a.create_empty(file_name=crit_res_result, zones=self.zones, matrix_names=queries["labels"])
                        else:
                            raise ValueError(
                                "Queries are inconsistent. 'Labels', 'elements' and 'type' need to have same dimensions"
                            )
                    else:
                        raise ValueError(
                            "Queries are inconsistent. It needs to contain the following elements: 'Labels', 'elements' and 'type'"
                        )
        else:
            a.create_empty(file_name=a.random_name(), zones=self.zones, matrix_names=["empty", "nothing"])

        a.computational_view()
        if len(a.matrix_view.shape[:]) == 2:
            a.matrix_view = a.matrix_view.reshape((self.zones, self.zones, 1))
        self.critical_links = {"save": save, "queries": queries, "results": a}

    def setSavePathFile(self, save=False, path_result=None):
        # Fields: Origin, Node, Predecessor
        # Number of records: Origins * Nodes
        a = AequilibraEData()
        d1 = max(1, self.zones)
        d2 = 1
        memory_mode = True

        if save:
            if path_result is None:
                warnings.warn("Path file not set properly. Need to specify output file too")
            else:
                # This is the only place where we keep 32bits, as going 64 bits would explode the file size
                if self.nodes > 0 and self.zones > 0:
                    d1 = self.zones
                    d2 = self.nodes
                    memory_mode = False

        a.create_empty(
            file_path=path_result,
            entries=d1 * d2,
            field_names=["origin", "node", "predecessor", "connector"],
            data_types=[np.uint32, np.uint32, np.uint32, np.uint32],
            memory_mode=memory_mode,
        )

        self.path_file = {"save": save, "results": a}

    def save_to_disk(self, file_name=None, output="loads"):
        """ Function to write to disk all outputs computed during assignment
    Args:
        output: 'loads', for writing the link loads
                'path_file', for writing the path file to a format different than the native binary

        file_name: Name of the file, with extension. Valid extensions are:
                   1. 'aed', for AequilibraE datasets. (Nearly zero overhead)
                   2. 'csv', for comma-separated files
                   3. 'sqlite' for sqlite databases
    Returns:
        Nothing"""

        if file_name is None:
            # memory = True
            file_type = "MEMORY"
        else:
            # memory = False
            fname, file_type = os.path.splitext(file_name.upper())

        fields = []
        if output == "loads":
            table_name = "link_loads"
            for n in self.classes["names"]:
                fields.extend([n + "_ab", n + "_ba", n + "_tot"])
            types = [np.float64] * len(fields)

            if file_type == ".AED":
                aed_file_name = file_name
                memory_mode = False
            else:
                memory_mode = True
                aed_file_name = None

            entries = int(np.unique(self.lids).shape[0])
            res = AequilibraEData()
            res.create_empty(
                file_path=aed_file_name, memory_mode=memory_mode, entries=entries, field_names=fields, data_types=types
            )

            res.index[:] = np.unique(self.lids)[:]

            indexing = np.zeros(int(self.lids.max()) + 1, np.uint64)
            indexing[res.index[:]] = np.arange(entries)

            # Indices of links BA and AB
            ABs = self.direcs > 0
            BAs = self.direcs < 0
            ab_ids = indexing[self.lids[ABs]]
            ba_ids = indexing[self.lids[BAs]]

            # Link flows
            link_flows = self.link_loads[:, :]
            for i, n in enumerate(self.classes["names"]):
                # AB Flows
                res.data[n + "_ab"][ab_ids] = np.nan_to_num(link_flows[ABs, i])
                # BA Flows
                res.data[n + "_ba"][ba_ids] = np.nan_to_num(link_flows[BAs, i])

                # Tot Flow
                res.data[n + "_tot"] = res.data[n + "_ab"] + res.data[n + "_ba"]

            # Save to disk
            if file_name is None:
                return res
            else:
                if file_type != "aed":
                    res.export(file_name, table_name=table_name)
                del res

        # TODO: Re-factor the exporting of the path file within the AequilibraeData format
        elif output == "path_file":
            pass



import os
import copy
import csv
import numpy as np

from .agency import Agency
from .calendar_dates import CalendarDates
from .stop import Stop
from .route import Route
from .parse_csv import parse_csv


class GTFS:
    """
     Reader for GTFS (from https://developers.google.com/transit/gtfs/reference/)

     .

     Objective
     _________
     To provide a memory container for GTFS that can be:
        * Passed to transit assignment algorithms in memory
        * Edited and saved back to disk
        * Displayed in a GIS environment
    """

    def __init__(self):
        self.database = None
        self.source_folder = None
        self.agency = Agency()
        self.trips = None
        self.num_routes = None
        self.routes = {}
        self.stops = {}
        self.calendar_dates = {}
        self.available_files = {}
        self.shapes = {}
        self.schedule_exceptions = None

    def load_from_file(self, file_path, save_db=False, memory_db=False):
        pass

    def load(self, path_to_folder):
        self.source_folder = path_to_folder

        self.load_agency()
        self.load_stops()
        self.load_routes()
        self.load_trips()
        self.load_stop_times()
        self.load_calendar()
        self.load_calendar_dates()
        self.load_shapes()

        self.get_routes_shapes()

    def load_agency(self) -> None:
        agency_file = os.path.join(self.source_folder, "agency.txt")
        self.available_files["agency.txt"] = True
        data = parse_csv(agency_file)
        if not len(data):
            return
        # TODO: Transfer to the database style
        self.agency.email = str(data["agency_id"][0])
        self.agency.name = str(data["agency_name"][0])
        self.agency.url = str(data["agency_url"][0])
        self.agency.timezone = str(data["agency_timezone"][0])
        self.agency.lang = str(data["agency_lang"][0])
        self.agency.phone = str(data["agency_phone"][0])
        del data

    def load_stops(self):
        stops_file = os.path.join(self.source_folder, "stops.txt")
        self.available_files["stops.txt"] = True
        data = parse_csv(stops_file)

        # Iterate over all the stops and puts them in the stops dictionary
        for i in range(data.shape[0]):
            stop = Stop()
            # Required fields
            stop.id = data["stop_id"][i]
            stop.name = data["stop_name"][i]
            stop.lat = data["stop_lat"][i]
            stop.lon = data["stop_lon"][i]

            # optional fields
            available_fields = data.dtype.names
            if "stop_code" in available_fields:
                stop.code = data["stop_code"][i]
            if "stop_desc" in available_fields:
                stop.desc = data["stop_desc"][i]
            if "zone_id" in available_fields:
                stop.zone_id = data["zone_id"][i]
            if "stop_url" in available_fields:
                stop.url = data["stop_url"][i]
            if "zone_id" in available_fields:
                stop.zone_id = data["zone_id"][i]
            if "location_type" in available_fields:
                stop.location_type = data["location_type"][i]
            if "parent_station" in available_fields:
                stop.parent_station = data["parent_station"][i]
            if "timezone" in available_fields:
                stop.timezone = data["timezone"][i]
            if "wheelchair_boarding" in available_fields:
                stop.wheelchair_boarding = data["wheelchair_boarding"][i]

            self.stops[stop.id] = stop
        del data

    def load_routes(self):
        routes_file = os.path.join(self.source_folder, "routes.txt")
        self.available_files["routes.txt"] = True
        data = parse_csv(routes_file)

        # Iterate over all the stops and puts them in the stops dictionary
        for i in range(data.shape[0]):
            r = Route()
            # Required fields
            r.id = data["route_id"][i]
            r.short_name = data["route_short_name"][i]
            r.long_name = data["route_long_name"][i]
            r.type = data["route_type"][i]

            # optional fields
            available_fields = data.dtype.names
            if "agency_id" in available_fields:
                r.agency_id = data["agency_id"][i]
            if "route_desc" in available_fields:
                r.desc = data["route_desc"][i]
            if "route_url" in available_fields:
                r.url = data["route_url"][i]
            if "route_color" in available_fields:
                r.color = data["route_color"][i]
            if "route_text_color" in available_fields:
                r.text_color = data["route_text_color"][i]
            if "route_sort_order" in available_fields:
                r.sort_order = data["route_sort_order"][i]
            self.routes[r.id] = r

        del data

    def load_trips(self):
        trips_file = os.path.join(self.source_folder, "trips.txt")
        self.available_files["trips.txt"] = True

        self.trips = parse_csv(trips_file)

    def load_stop_times(self):
        stop_times_file = os.path.join(self.source_folder, "stop_times.txt")
        self.available_files["stop_times.txt"] = True
        self.stop_times = parse_csv(stop_times_file)

    def load_calendar(self):
        pass

    def load_calendar_dates(self):
        agency_file = os.path.join(self.source_folder, "calendar_dates.txt")
        if not os.path.isfile(agency_file):
            self.available_files["calendar_dates.txt"] = False
            return

        self.available_files["calendar_dates.txt"] = True
        data = parse_csv(agency_file)
        all_exceptions = []
        for i in range(data.shape[0]):
            cd = CalendarDates()
            # Required fields
            cd.service_id = data["service_id"][i]
            cd.date = data["date"][i]
            cd.exception_type = data["exception_type"][i]
            all_exceptions.append(cd.service_id)
            self.calendar_dates[i] = cd
        self.schedule_exceptions = set(all_exceptions)
        del all_exceptions
        del data

    def load_shapes(self):
        # TODO: Add the info from field "shape_dist_traveled"
        shapes_file = os.path.join(self.source_folder, "shapes.txt")
        if not os.path.isfile(shapes_file):
            self.available_files["shapes.txt"] = False
            return

        self.available_files["shapes.txt"] = True
        data = parse_csv(shapes_file)

        all_shapes = list(np.unique(data["shape_id"]))

        for shp in all_shapes:
            trace = data[data["shape_id"] == shp]
            trace = np.sort(trace, order=["shape_pt_sequence"])
            coords = np.core.defchararray.add(trace["shape_pt_lon"].astype(str), " ")
            coords = np.core.defchararray.add(coords, trace["shape_pt_lat"].astype(str))
            coords = ", ".join(list(coords))
            self.shapes[shp] = '"LINESTRING(' + coords + ')"'

    def get_routes_shapes(self):
        for rt in self.routes.keys():
            trips = self.trips[self.trips["route_id"] == rt]["shape_id"]
            if self.available_files["shapes.txt"]:
                self.routes[rt].shapes = {t: self.shapes[t] for t in trips}
            # else:
            #     for t in trips:
            #         stop_times = self.stop_times[self.stop_times["trip_id"] == t]

    def get_routes_stops(self):
        pass



from unittest import TestCase

import numpy as np

from aequilibrae.paths import Graph
from aequilibrae.paths import NetworkSkimming
from aequilibrae.paths import skimming_single_origin
from aequilibrae.paths.results import SkimResults

# Adds the folder with the data to the path and collects the paths to the files
from ...data import test_graph


class MultiThreadedNetworkSkimming:
    def __init__(self):
        self.predecessors = None  # The predecessors for each node in the graph
        # holds the skims for all nodes in the network (during path finding)
        self.temporary_skims = None
        # Keeps the order in which the nodes were reached for the cascading network loading
        self.reached_first = None
        self.connectors = None  # The previous link for each node in the tree
        # holds the b_nodes in case of flows through centroid connectors are blocked
        self.temp_b_nodes = None

    # In case we want to do by hand, we can prepare each method individually
    def prepare(self, graph, results):
        itype = graph.default_types("int")
        ftype = graph.default_types("float")
        self.predecessors = np.zeros((results.nodes, results.cores), dtype=itype)
        self.temporary_skims = np.zeros((results.nodes, results.num_skims, results.cores), dtype=ftype)
        self.reached_first = np.zeros((results.nodes, results.cores), dtype=itype)
        self.connectors = np.zeros((results.nodes, results.cores), dtype=itype)
        self.temp_b_nodes = np.zeros((graph.b_node.shape[0], results.cores), dtype=itype)
        for i in range(results.cores):
            self.temp_b_nodes[:, i] = graph.b_node[:]


class TestNetwork_skimming(TestCase):
    def test_network_skimming(self):
        # graph
        g = Graph()
        g.load_from_disk(test_graph)
        g.set_graph(cost_field="distance", skim_fields=None)
        # None implies that only the cost field will be skimmed

        # skimming results
        res = SkimResults()
        res.prepare(g)

        aux_res = MultiThreadedNetworkSkimming()
        aux_res.prepare(g, res)
        _ = skimming_single_origin(26, g, res, aux_res, 0)

        skm = NetworkSkimming(g, res)
        skm.execute()

        tot = np.nanmax(res.skims.distance[:, :])

        if tot > 10e10:
            self.fail("Skimming was not successful. At least one np.inf returned.")

        if skm.report:
            self.fail("Skimming returned an error:" + str(skm.report))



"""
=============
Trip distribution code
=============

Imports procedures for trip distribution

"""

__author__ = "Pedro Camargo ($Author: Pedro Camargo $)"
__version__ = "1.1"
__revision__ = "$Revision: 2 $"
__date__ = "$Date: 2017-10-05$"

from .ipf import Ipf
from .gravity_application import GravityApplication
from .gravity_calibration import GravityCalibration
from .synthetic_gravity_model import SyntheticGravityModel



from .gtfs import GTFS
from .route import Route
from .agency import Agency
from .stop import Stop
from .trip import Trip
from .calendar_dates import CalendarDates
from .gtfs_sqlite_db import create_gtfsdb



"""
-----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Network skimming
 Purpose:    Implement skimming algorithms based on Cython's path finding and skimming

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camrgo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    2017-07-03
 Updated:    2017-05-07
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
-----------------------------------------------------------------------------------------------------------
 """

import importlib
import sys
import threading
from multiprocessing.dummy import Pool as ThreadPool

from .AoN import skimming_single_origin
from .multi_threaded_skimming import MultiThreadedNetworkSkimming
from ..utils import WorkerThread

have_pyqt5 = importlib.util.find_spec("PyQt5")
if have_pyqt5 is None:
    pyqt = False
else:
    from PyQt5.QtCore import pyqtSignal as SIGNAL

    pyqt = True

sys.dont_write_bytecode = True


class NetworkSkimming(WorkerThread):

    skimming = SIGNAL(object)

    def __init__(self, graph, results, origins=None):
        WorkerThread.__init__(self, None)

        self.origins = origins
        self.graph = graph
        self.results = results
        self.aux_res = MultiThreadedNetworkSkimming()
        self.report = []
        self.cumulative = 0

        if results.__graph_id__ != graph.__id__:
            raise ValueError("Results object not prepared. Use --> results.prepare(graph)")

        if results.__graph_id__ is None:
            raise ValueError("The results object was not prepared. Use results.prepare(graph)")

        elif results.__graph_id__ != graph.__id__:
            raise ValueError("The results object was prepared for a different graph")

    def doWork(self):
        self.execute()

    def execute(self):
        if pyqt:
            self.skimming.emit(["zones finalized", 0])

        self.aux_res.prepare(self.graph, self.results)

        pool = ThreadPool(self.results.cores)
        all_threads = {"count": 0}
        for orig in list(self.graph.centroids):
            i = int(self.graph.nodes_to_indices[orig])
            if i >= self.graph.nodes_to_indices.shape[0]:
                self.report.append("Centroid " + str(orig) + " is beyond the domain of the graph")
            elif self.graph.fs[int(i)] == self.graph.fs[int(i) + 1]:
                self.report.append("Centroid " + str(orig) + " does not exist in the graph")
            else:
                pool.apply_async(self.func_assig_thread, args=(orig, all_threads))
        pool.close()
        pool.join()

        if pyqt:
            self.skimming.emit(["text skimming", "Saving Outputs"])
            self.skimming.emit(["finished_threaded_procedure", None])

    def func_assig_thread(self, O, all_threads):
        if threading.get_ident() in all_threads:
            th = all_threads[threading.get_ident()]
        else:
            all_threads[threading.get_ident()] = all_threads["count"]
            th = all_threads["count"]
            all_threads["count"] += 1
        x = skimming_single_origin(O, self.graph, self.results, self.aux_res, th)
        self.cumulative += 1
        if x != O:
            self.report.append(x)
        if pyqt:
            self.skimming.emit(["zones finalized", self.cumulative])
            txt = str(self.cumulative) + " / " + str(self.matrix.zones)
            self.skimming.emit(["text skimming", txt])



class Stop:
    """
    Represents each one of the physical stops in a GTFS dataset (from https://developers.google.com/transit/gtfs/reference/)

    Fields
    ______

    * **id** `(stop_id)` **Required** - The stop_id field contains an ID that uniquely identifies a stop, station, or station entrance. Multiple routes may use the same stop. The stop_id is used by systems as an internal identifier of this record (e.g., primary key in database), and therefore the stop_id must be dataset unique.

    * **code** `(stop_code)` **Optional** - The stop_code field contains short text or a number that uniquely identifies the stop for passengers. Stop codes are often used in phone-based transit information systems or printed on stop signage to make it easier for riders to get a stop schedule or real-time arrival information for a particular stop. The stop_code field contains short text or a number that uniquely identifies the stop for passengers. The stop_code can be the same as stop_id if it is passenger-facing. This field should be left blank for stops without a code presented to passengers.

    * **name** `(stop_name)` **Required** - The stop_name field contains the name of a stop, station, or station entrance. Please use a name that people will understand in the local and tourist vernacular.

    * **desc**  `(stop_desc)` **Optional** - The stop_desc field contains a description of a stop. Please provide useful, quality information. Do not simply duplicate the name of the stop.

    * **lat** `(stop_lat)` **Required** - The stop_lat field contains the latitude of a stop, station, or station entrance. The field value must be a valid WGS 84 latitude.

    * **lon** `(stop_lon)` **Required** - The stop_lon field contains the longitude of a stop, station, or station entrance. The field value must be a valid WGS 84 longitude value from -180 to 180.

    * **zone_id** `(zone_id)` **Optional** - The zone_id field defines the fare zone for a stop ID. Zone IDs are required if you want to provide fare information using fare_rules.txt. If this stop ID represents a station, the zone ID is ignored.

    * **url** `(stop_url)` **Optional** - The stop_url field contains the URL of a web page about a particular stop. This should be different from the agency_url and the route_url fields. The value must be a fully qualified URL that includes http:// or https://, and any special characters in the URL must be correctly escaped. See http://www.w3.org/Addressing/URL/4_URI_Recommentations.html for a description of how to create fully qualified URL values.

    * **location_type** `(location_type)` **Optional** - The location_type field identifies whether this stop ID represents a stop, station, or station entrance. If no location type is specified, or the location_type is blank, stop IDs are treated as stops. Stations may have different properties from stops when they are represented on a map or used in trip planning. The location type field can have the following values:

            - 0 or blank - Stop. A location where passengers board or disembark from a transit vehicle.
            - 1 - Station. A physical structure or area that contains one or more stop.
            - 2 - Station Entrance/Exit. A location where passengers can enter or exit a station from the street. The stop entry must also specify a parent_station value referencing the stop ID of the parent station for the entrance.
    * **parent_station** `(parent_station)` **Optional** - For stops that are physically located inside stations, the parent_station field identifies the station associated with the stop. To use this field, stops.txt must also contain a row where this stop ID is assigned location type=1.

            This stop ID represents...	This entry's location type...	This entry's parent_station field contains...
            A stop located inside a station.	0 or blank	The stop ID of the station where this stop is located. The stop referenced by parent_station must have location_type=1.
            A stop located outside a station.	0 or blank	A blank value. The parent_station field doesn't apply to this stop.
            A station.	1	A blank value. Stations can't contain other stations.

    * **timezone** `(stop_timezone)` **Optional** - The stop_timezone field contains the timezone in which this stop, station, or station entrance is located. Please refer to Wikipedia List of Timezones for a list of valid values. If omitted, the stop should be assumed to be located in the timezone specified by agency_timezone in agency.txt. When a stop has a parent station, the stop is considered to be in the timezone specified by the parent station's stop_timezone value. If the parent has no stop_timezone value, the stops that belong to that station are assumed to be in the timezone specified by agency_timezone, even if the stops have their own stop_timezone values. In other words, if a given stop has a parent_station value, any stop_timezone value specified for that stop must be ignored. Even if stop_timezone values are provided in stops.txt, the times in stop_times.txt should continue to be specified as time since midnight in the timezone specified by agency_timezone in agency.txt. This ensures that the time values in a trip always increase over the course of a trip, regardless of which timezones the trip crosses.

    * **wheelchair_boarding** `(wheelchair_boarding)` **Optional** - The wheelchair_boarding field identifies whether wheelchair boardings are possible from the specified stop, station, or station entrance. The field can have the following values:

            - 0 (or empty) - indicates that there is no accessibility information for the stop
            - 1 - indicates that at least some vehicles at this stop can be boarded by a rider in a wheelchair
            - 2 - wheelchair boarding is not possible at this stop
        When a stop is part of a larger station complex, as indicated by a stop with a parent_station value, the stop's wheelchair_boarding field has the following additional semantics:
            -  0 (or empty) - the stop will inherit its wheelchair_boarding value from the parent station, if specified in the parent
            - 1 - there exists some accessible path from outside the station to the specific stop / platform
            - 2 - there exists no accessible path from outside the station to the specific stop / platform
        For station entrances, the wheelchair_boarding field has the following additional semantics:
            - 0 (or empty) - the station entrance will inherit its wheelchair_boarding value from the parent station, if specified in the parent
            - 1 - the station entrance is wheelchair accessible (e.g. an elevator is available to platforms if they are not at-grade)
            - 2 - there exists no accessible path from the entrance to station platforms
    """

    def __init__(self):
        """
        Initializes the class with members corresponding to all fields in the GTFS specification. See Stop class
        documentation
        """

        self.id = None
        self.code = ""
        self.name = None
        self.desc = ""
        self.lat = None
        self.lon = None
        self.zone_id = None
        self.url = None
        self.location_type = 0
        self.parent_station = None
        self.timezone = None
        self.wheelchair_boarding = 0



"""
path computation related code
"""

__author__ = "Pedro Camargo ($Author: Pedro Camargo $)"
__version__ = "0.4.0"
__revision__ = "$Revision: 2 $"
__date__ = "$Date: 2017-02-25$"

import warnings

try:
    from .AoN import one_to_all, skimming_single_origin, path_computation, VERSION_COMPILED, update_path_trace
    from .results import *
    from .multi_threaded_aon import MultiThreadedAoN
    from .multi_threaded_skimming import MultiThreadedNetworkSkimming
    from .network_skimming import NetworkSkimming
    from .all_or_nothing import allOrNothing
except ImportError as e:
    warnings.warn("The AoN extension has not been compiled. {}".format(e.name))
from .graph import Graph
from .__version__ import binary_version, release_name, minor_version, release_version



import glob
import os
import tempfile


def clean_after_tests():
    p = tempfile.gettempdir() + "/aequilibrae_*"
    for f in glob.glob(p):
        try:
            os.unlink(f)
        except Exception as err:
            print(err.__str__())


clean_after_tests()



from unittest import TestCase

import numpy as np

from aequilibrae.distribution import GravityCalibration
from aequilibrae.matrix import AequilibraeMatrix

zones = 100

# Impedance matrix_procedures
name_test = AequilibraeMatrix().random_name()

args = {"file_name": name_test, "zones": zones, "matrix_names": ["impedance"]}

impedance = AequilibraeMatrix()
impedance.create_empty(**args)
impedance.impedance[:, :] = np.random.rand(zones, zones)[:, :] * 1000
impedance.index[:] = np.arange(impedance.zones) + 100
impedance.computational_view(["impedance"])

args["matrix_names"] = ["base_matrix"]
args["file_name"] = AequilibraeMatrix().random_name()
matrix = AequilibraeMatrix()
matrix.create_empty(**args)
matrix.base_matrix[:, :] = np.random.rand(zones, zones)[:, :] * 1000
matrix.index[:] = np.arange(matrix.zones) + 100
matrix.computational_view(["base_matrix"])


class TestGravityCalibration(TestCase):
    def test_calibrate(self):
        args = {"impedance": impedance, "matrix": matrix, "function": "power", "nan_to_zero": False}

        distributed_matrix = GravityCalibration(**args)
        distributed_matrix.calibrate()
        if distributed_matrix.gap > 0.0001:
            self.fail("Calibration did not converge")

    def test_check_inputs(self):
        # self.fail()
        pass

    def test_apply_gravity(self):
        # self.fail()
        pass

    def test_get_parameters(self):
        # self.fail()
        pass



"""
-----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:      Synthetic gravity trip distribution model application
 Purpose:    Implementing the algorithms to apply trip distribution.
                  Implemented: Synthetic gravity with power, exponential and gamma functions
                  Still missing: 2nd stage: Friction factors

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    2016-09-30
 Updated:    2017-08-11
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------
 """
# The procedures implemented in this code are some of those suggested in
# Modelling Transport, 4th Edition
# Ortuzar and Willumsen, Wiley 2011
# The referred authors have no responsibility over this work, of course
from .ipf import Ipf
from .synthetic_gravity_model import SyntheticGravityModel
from ..matrix import AequilibraeMatrix, AequilibraEData
from ..parameters import Parameters

import sys
import numpy as np
import os
import tempfile
from time import clock
import glob
import logging

sys.dont_write_bytecode = True


class GravityApplication:
    """"
    Model is an instance of SyntheticGravityModel class
    Impedance is an instance of AequilibraEMatrix
    Row and Column vectors are instances of AequilibraEData
    """

    def __init__(self, **kwargs):

        self.__required_parameters = ["max trip length"]
        self.__required_model = ["function", "parameters"]

        self.parameters = kwargs.get("parameters", self.get_parameters())

        self.rows = kwargs.get("rows")
        self.row_field = kwargs.get("row_field", None)

        self.columns = kwargs.get("columns")
        self.column_field = kwargs.get("column_field", None)

        self.impedance = kwargs.get("impedance")
        self.model = kwargs.get("model")
        self.core_name = kwargs.get("output_core", "gravity")
        self.output_name = kwargs.get("output", AequilibraeMatrix().random_name())
        self.nan_as_zero = kwargs.get("nan_as_zero", False)
        self.output = None
        self.gap = np.inf
        self.logger = logging.getLogger("aequilibrae")

    def apply(self):
        self.check_data()
        t = clock()
        max_cost = self.parameters["max trip length"]
        # We create the output
        self.output = self.impedance.copy(self.output_name, cores=self.impedance.view_names, names=[self.core_name])
        self.output.computational_view([self.core_name])
        if self.nan_as_zero:
            self.output.matrix_view[:, :] = np.nan_to_num(self.output.matrix_view)[:, :]

        # We apply the function
        self.apply_function()

        # We zero those cells that have a trip length above the limit
        if max_cost > 0:
            a = (self.output.matrix_view[:, :] < max_cost).astype(int)
            self.output.matrix_view[:, :] = a * self.output.matrix_view[:, :]

        # We adjust the total of the self.output
        total_factor = np.nansum(self.rows.data[self.row_field]) / np.nansum(self.output.matrix_view[:, :])
        self.output.matrix_view[:, :] = self.output.matrix_view[:, :] * total_factor

        # And adjust with a fratar
        ipf = Ipf(
            matrix=self.output,
            rows=self.rows,
            columns=self.columns,
            column_field=self.column_field,
            row_field=self.row_field,
            nan_as_zero=self.nan_as_zero,
        )

        # We use the model application parameters in case they were provided
        # not the standard way of using this tool)
        for p in ipf.parameters:
            if p in self.parameters:
                ipf.parameters[p] = self.parameters[p]

        # apply fratar
        ipf.fit()
        self.output = ipf.output
        self.gap = ipf.gap

        q = ipf.report.pop(0)
        for q in ipf.report:
            self.report.append(q)

        self.report.append("")
        self.report.append("")

        self.report.append("Total of matrix: " + "{:15,.4f}".format(float(np.nansum(self.output.matrix_view))))
        self.report.append(
            "Intrazonal flow: " + "{:15,.4f}".format(float(np.nansum(np.diagonal(self.output.matrix_view))))
        )
        self.report.append("Running time: " + str(round(clock() - t, 3)))

        for i in glob.glob(tempfile.gettempdir() + "*.aem"):
            try:
                os.unlink(i)
            except PermissionError as err:
                self.logger.warning("Could not remove " + err.filename)

    def get_parameters(self):
        par = Parameters().parameters
        para = par["distribution"]["ipf"].copy()
        para.update(par["distribution"]["gravity"])
        return para

    def check_data(self):
        self.report = ["  #####    GRAVITY APPLICATION    #####  ", ""]

        if not isinstance(self.model, SyntheticGravityModel):
            self.error_free = False
            raise TypeError("Model is not an instance of SyntheticGravityModel")

        self.report.append("Model specification:")
        self.report.append("    Function: " + self.model.function)
        if self.model.alpha is not None:
            self.report.append("    alpha: " + str(self.model.alpha))

        if self.model.beta is not None:
            self.report.append("    beta: " + str(self.model.beta))

        self.report.append("")

        # check dimensions
        # check data types
        if not isinstance(self.rows, AequilibraEData):
            raise TypeError("Row vector needs to be an instance of AequilibraEData")

        if not isinstance(self.columns, AequilibraEData):
            raise TypeError("Column vector needs to be an instance of AequilibraEData")

        if not isinstance(self.impedance, AequilibraeMatrix):
            raise TypeError("Impedance matrix needs to be an instance of AequilibraEMatrix")

        # Check data dimensions
        if not np.array_equal(self.rows.index, self.columns.index):
            raise ValueError("Indices from row vector do not match those from column vector")

        if not np.array_equal(self.impedance.index, self.columns.index):
            raise ValueError("Indices from vectors do not match those from seed matrix")

        # Check if matrix was set for computation
        if self.impedance.matrix_view is None:
            raise ValueError("Matrix needs to be set for computation")
        else:
            if len(self.impedance.matrix_view.shape[:]) > 2:
                raise ValueError("Matrix' computational view needs to be set for a single matrix core")

        # check balancing:
        sum_rows = np.nansum(self.rows.data[self.row_field])
        sum_cols = np.nansum(self.columns.data[self.column_field])
        if abs(sum_rows - sum_cols) > self.parameters["balancing tolerance"]:
            raise ValueError("Vectors are not balanced")
        else:
            # guarantees that they are precisely balanced
            self.columns.data[self.column_field][:] = self.columns.data[self.column_field][:] * (sum_rows / sum_cols)

        self.check_parameters()

    def check_parameters(self):
        # Check if parameters are configured properly
        for p in self.__required_parameters:
            if p not in self.parameters:
                self.error = "Parameters error. It needs to be a dictionary with the following keys: "
                for t in self.__required_parameters:
                    self.error = self.error + t + ", "
                break

    def apply_function(self):
        self.core_name = self.output.view_names[0]
        for i in range(self.rows.entries):
            p = self.rows.data[self.row_field][i]
            a = self.columns.data[self.column_field][:]

            if self.model.function == "EXPO":
                self.output.matrix_view[i, :] = np.exp(-self.model.beta * self.impedance.matrix_view[i, :]) * p * a

            elif self.model.function == "POWER":
                # self.output.matrices[self.core_name][i, :] = (np.power(self.impedance.matrix_view[i, :, 0], - self.model.alpha) * p * a)[:]
                self.output.matrix_view[i, :] = (np.power(self.impedance.matrix_view[i, :], -self.model.alpha) * p * a)[
                    :
                ]
            elif self.model.function == "GAMMA":
                self.output.matrix_view[i, :] = (
                    np.power(self.impedance.matrix_view[i, :], self.model.alpha)
                    * np.exp(-self.model.beta * self.impedance.matrix_view[i, :])
                    * p
                    * a
                )[:]

        # Deals with infinite and NaNs
        infinite = np.isinf(self.output.matrix_view[:, :]).astype(int)
        non_inf = np.ones_like(self.output.matrix_view[:, :]) - infinite
        self.output.matrix_view[:, :] = self.output.matrix_view[:, :] * non_inf
        np.nan_to_num(self.output.matrix_view[:, :])



"""
 -----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Iterative proportional fitting
 Purpose:    Implement Iterative proportional fitting

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    29/09/2016
 Updated:    11/08/2017
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------
"""
import os
import sys
from time import clock

import numpy as np
import yaml

from ..matrix import AequilibraEData
from ..matrix import AequilibraeMatrix

sys.dont_write_bytecode = True


class Ipf:
    def __init__(self, **kwargs):

        self.parameters = kwargs.get("parameters", self.get_parameters("ipf"))

        # Seed matrix
        self.matrix = kwargs.get("matrix", None)

        # NaN as zero
        self.nan_as_zero = kwargs.get("nan_as_zero", True)

        # row vector
        self.rows = kwargs.get("rows", None)
        self.row_field = kwargs.get("row_field", None)
        self.output_name = kwargs.get("output", AequilibraeMatrix().random_name())

        # Column vector
        self.columns = kwargs.get("columns", None)
        self.column_field = kwargs.get("column_field", None)

        self.output = None
        self.error = None
        self.__required_parameters = ["convergence level", "max iterations", "balancing tolerance"]
        self.error_free = True
        self.report = ["  #####    IPF computation    #####  ", ""]
        self.gap = None

    def check_data(self):
        self.error = None
        self.check_parameters()

        # check data types
        if not isinstance(self.rows, AequilibraEData):
            raise TypeError("Row vector needs to be an instance of AequilibraEData")

        if not isinstance(self.columns, AequilibraEData):
            raise TypeError("Column vector needs to be an instance of AequilibraEData")

        if not isinstance(self.matrix, AequilibraeMatrix):
            raise TypeError("Seed matrix needs to be an instance of AequilibraEMatrix")

        # Check data type
        if not np.issubdtype(self.matrix.dtype, np.floating):
            raise ValueError("Seed matrix need to be a float type")

        if not np.issubdtype(self.rows.data[self.row_field].dtype, np.floating):
            raise ValueError("production/rows vector must be a float type")

        if not np.issubdtype(self.columns.data[self.column_field].dtype, np.floating):
            raise ValueError("Attraction/columns vector must be a float type")

        # Check data dimensions
        if not np.array_equal(self.rows.index, self.columns.index):
            raise ValueError("Indices from row vector do not match those from column vector")

        if not np.array_equal(self.matrix.index, self.columns.index):
            raise ValueError("Indices from vectors do not match those from seed matrix")

        # Check if matrix was set for computation
        if self.matrix.matrix_view is None:
            raise ValueError("Matrix needs to be set for computation")
        else:
            if len(self.matrix.matrix_view.shape[:]) > 2:
                raise ValueError("Matrix' computational view needs to be set for a single matrix core")

        if self.error is None:
            # check balancing:
            sum_rows = np.nansum(self.rows.data[self.row_field])
            sum_cols = np.nansum(self.columns.data[self.column_field])
            if abs(sum_rows - sum_cols) > self.parameters["balancing tolerance"]:
                self.error = "Vectors are not balanced"
            else:
                # guarantees that they are precisely balanced
                self.columns.data[self.column_field][:] = self.columns.data[self.column_field][:] * (
                    sum_rows / sum_cols
                )

        if self.error is not None:
            self.error_free = False

    def check_parameters(self):
        for i in self.__required_parameters:
            if i not in self.parameters:
                self.error = "Parameters error. It needs to be a dictionary with the following keys: "
                for t in self.__required_parameters:
                    self.error = self.error + t + ", "
                break

    def fit(self):
        t = clock()
        self.check_data()
        if self.error_free:
            max_iter = self.parameters["max iterations"]
            conv_criteria = self.parameters["convergence level"]

            self.output = self.matrix.copy(self.output_name)
            if self.nan_as_zero:
                self.output.matrix_view[:, :] = np.nan_to_num(self.output.matrix_view)[:, :]

            rows = self.rows.data[self.row_field]
            columns = self.columns.data[self.column_field]
            tot_matrix = np.nansum(self.output.matrix_view[:, :])

            # Reporting
            self.report.append("Target convergence criteria: " + str(conv_criteria))
            self.report.append("Maximum iterations: " + str(max_iter))
            self.report.append("")
            self.report.append("Rows:" + str(self.rows.entries))
            self.report.append("Columns: " + str(self.columns.entries))

            self.report.append("Total of seed matrix: " + "{:28,.4f}".format(float(tot_matrix)))
            self.report.append("Total of target vectors: " + "{:25,.4f}".format(float(np.nansum(rows))))
            self.report.append("")
            self.report.append("Iteration,   Convergence")
            self.gap = conv_criteria + 1

            iter = 0
            while self.gap > conv_criteria and iter < max_iter:
                iter += 1
                # computes factors for zones
                marg_rows = self.tot_rows(self.output.matrix_view[:, :])
                row_factor = self.factor(marg_rows, rows)
                # applies factor
                self.output.matrix_view[:, :] = np.transpose(
                    np.transpose(self.output.matrix_view[:, :]) * np.transpose(row_factor)
                )[:, :]

                # computes factors for columns
                marg_cols = self.tot_columns(self.output.matrix_view[:, :])
                column_factor = self.factor(marg_cols, columns)

                # applies factor
                self.output.matrix_view[:, :] = self.output.matrix_view[:, :] * column_factor

                # increments iterarions and computes errors
                self.gap = max(
                    abs(1 - np.min(row_factor)),
                    abs(np.max(row_factor) - 1),
                    abs(1 - np.min(column_factor)),
                    abs(np.max(column_factor) - 1),
                )

                self.report.append(str(iter) + "   ,   " + str("{:4,.10f}".format(float(np.nansum(self.gap)))))

            self.report.append("")
            self.report.append("Running time: " + str("{:4,.3f}".format(clock() - t)) + "s")

    def tot_rows(self, matrix):
        return np.nansum(matrix, axis=1)

    def tot_columns(self, matrix):
        return np.nansum(matrix, axis=0)

    def factor(self, marginals, targets):
        f = np.divide(targets, marginals)  # We compute the factors
        f[f == np.NINF] = 1  # And treat the errors
        return f

    def get_parameters(self, model):
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
        with open(path + "/parameters.yml", "r") as yml:
            path = yaml.safe_load(yml)
        return path["distribution"][model]



"""
 -----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Worker thread
 Purpose:    Implements worker thread

 Original Author:  UNKNOWN. COPIED FROM STACKOVERFLOW BUT CAN'T REMEMBER EXACTLY WHERE
 Contributors:
 Last edited by: Pedro Camargo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    2014-03-19
 Updated:    2018-08-20
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------
 """

from PyQt5.QtCore import QThread
from PyQt5.QtCore import pyqtSignal


class WorkerThread(QThread):
    jobFinished = pyqtSignal(object)

    def __init__(self, parentThread):
        super().__init__()
        QThread.__init__(self, parentThread)

    def run(self):
        self.running = True
        success = self.doWork()
        self.jobFinished.emit(success)

    def stop(self):
        self.running = False
        pass

    def doWork(self):
        return True

    def cleanUp(self):
        pass
